package biz.capitalware.mqfm;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.jdom.JDOMException;

import biz.capitalware.mqfm.action.handlers.ProcessActions;
import biz.capitalware.mqfm.action.objects.*;
import biz.capitalware.mqfm.actions.*;
import biz.capitalware.mqfm.handlers.IO;
import biz.capitalware.mqfm.handlers.RunJob;
import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.mq.MQTrigger;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.MQFMException;
import biz.capitalware.mqfm.other.GeneralInfo;
import biz.capitalware.mqfm.xml.MQFMXml;

/**
 * This class will handle all MQ File Mover actions.
 *
 * @author Roger Lacroix, Capitalware Inc.
 * @return 0 for ok or 99 for failed.
 * @version 4.1.0
 * @license Apache 2 License
 */
public class MQFM
{
   private static final String               MQFM_XML_FILENAME = "mqfm.xml";
   //
   public  static final String               WHO_AM_I = "MQFM";
   public  static Logger                     logger = Logger.getLogger(WHO_AM_I);
   private static String                     xmlFilename = null;
   public  static String                     globalErrorText = null;
   public  static Hashtable<String,String>   userProperties = null;
   public  static Hashtable<String,FMObject> actionObjectValues = new Hashtable<String,FMObject>();
   public  static String                     processFileLock = null;
   public  static boolean                    processLock = false;
   public  static boolean                    useFileLocking = true;

   /**
    * The constructor
    */
   public MQFM()
   {
      super();
   }

   /**
    * Set the MQFM XML Workflow filename.
    * @param xmlFilename
    */
   public void setXmlFilename(String xmlFilename)
   {
      MQFM.xmlFilename = xmlFilename;
   }

   /**
    * @param args
    */
   public static void main(String[] args)
   {
      MQFM     mqfm = new MQFM();
      int      exitCode = 0;
      char     processType = Constants.ONE_PASS;
      boolean  windowsService = false;
      boolean  doWorkFlag = false;

      if (args.length > 3)
      {
         GeneralInfo.DumpHeader();
         MQFM.logger.fatal("Invalid number of command-line parameters");
         mqfm.usage();
         exitCode = 99;
         System.exit(exitCode);
      }
      else if (args.length == 0)
      {
         xmlFilename = MQFM_XML_FILENAME;  /* set a default value */
      }
      else if ("WindowsService".equals(args[0]))
      {
         windowsService = true;
         xmlFilename = mqfm.handleServiceStartup();
      }
      else
      {
         if (mqfm.isTrigger(args))
            processType = Constants.TRIGGERED;
         else
            xmlFilename = args[0];
      }

      /**
       * Check for process locking
       */
      if(args.length == 3)
      {
         processFileLock = args[1];
         File pfFile = new File(processFileLock);
         boolean invert = Boolean.parseBoolean(args[2].toLowerCase());

         if(pfFile.exists() && invert == true)
         {
            doWorkFlag = true;
         }
         else if(pfFile.exists() && invert == false)
         {
            doWorkFlag = false;
            MQFM.logger.info(WHO_AM_I + " is now ending as another instance is running.");
         }
         else if(!pfFile.exists() && invert == true)
         {
            doWorkFlag = false;
            MQFM.logger.info(WHO_AM_I + " is now ending a lock file does not exist and inverse locking is enabled.");
         }
         else
         {
            try 
            {
               boolean success = pfFile.createNewFile();
               if(success)
               {
                  MQFM.logger.info("Process lock file " + processFileLock + " created");
                  processLock = true;
                  doWorkFlag = true;
               }
               else
               {
                  doWorkFlag = false;
                  MQFM.logger.fatal("Cannot create file for process lock: " + processFileLock);
               }
            }
            catch (IOException e)
            {
               doWorkFlag = false;
               MQFM.logger.fatal( "IOException: " + e.getLocalizedMessage());
               if (e.getCause() != null)
                  MQFM.logger.fatal(e.getCause().toString());
            }
         }
      }
      else
      {
         doWorkFlag = true;
      }
      
      if(doWorkFlag)
      {
         Runtime.getRuntime().addShutdownHook(new ShutdownThread());
         exitCode = mqfm.doWork(processType, true);
      }

      MQFM.logger.info(WHO_AM_I + " is now ending.");

      System.exit(windowsService?0:exitCode);
   }

   /**
    * We have an MQFM XML file now it is time to do the work.
    * @param processType
    * @return exitCode
    */
   public int doWork(char processType, boolean headerFlag)
   {
      int   exitCode = 0;
      MQFMXml fmx;
      OnErrorObject onError = null;

      try
      {
         fmx = new MQFMXml();
         fmx.setLogger(logger);
         fmx.readXMLFile(xmlFilename);
         if (headerFlag)
            GeneralInfo.DumpHeader();
         MQFM.logger.info(WHO_AM_I + " has processed " + xmlFilename);
         onError = fmx.getOnError();
         userProperties = fmx.getProperties();
         if (headerFlag)
            GeneralInfo.GlobalValues(userProperties);

         exitCode = ProcessActions.doWork(fmx.getActions());

         MQFM.logger.info("");

         if (exitCode != 0)
            handleOnError(onError);
      }
      catch (JDOMException e)
      {
         MQFM.logger.fatal("JDOMException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
         globalErrorText = e.getLocalizedMessage();
         handleOnError(onError);
         exitCode = 99;
      }
      catch (FileNotFoundException e)
      {
         MQFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
         globalErrorText = e.getLocalizedMessage();
         handleOnError(onError);
         exitCode = 99;
      }
      catch (IOException e)
      {
         MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
         globalErrorText = e.getLocalizedMessage();
         handleOnError(onError);
         exitCode = 99;
      }
      catch (Exception e)
      {
         MQFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
         MQFM.globalErrorText = e.getLocalizedMessage();
         handleOnError(onError);
         exitCode = 99;
      }

      return exitCode;
   }

   /**
    * On error - perform either Execute or SendEmail action or both
    * @param oeo
    */
   private void handleOnError(OnErrorObject oeo)
   {
      MQFM.logger.fatal("MQFM Workflow XML file is abnormally terminating.");

      try
      {
         if (oeo != null)
         {
            if (oeo.eo != null)
            {
               MQFM.logger.fatal("OnError 'Execute Action' is now being performed.");
   
               RunJob rj = new RunJob();
               rj.launch(oeo.eo.xmlFile, globalErrorText);
            }
            
            if (oeo.seo != null)
            {
               MQFM.logger.fatal("OnError 'SendEmail Action' is now being performed.");
   
               SendEmailAction sea = new SendEmailAction(oeo.seo);
               sea.doWork();
            }
         }
      }
      catch (JDOMException e)
      {
         MQFM.logger.fatal("JDOMException: OnError: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
      }
      catch (FileNotFoundException e)
      {
         MQFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
      }
      catch (IOException e)
      {
         MQFM.logger.fatal("IOException: OnError: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
      }
   }

   /**
    * Was MQFM triggered?
    * @param args
    * @return boolean
    */
   private boolean isTrigger(String[] args)
   {
      boolean flag = false;

      if ( ( (args.length == 1) || (args.length == 2) ) &&
           (args[0].length() >= Constants.MQTMC_V1_LENGTH) &&
           (args[0].startsWith("TMC ")) )
      {
         try
         {
            MQTrigger mqt = new MQTrigger(args[0]);

            /* Are there any parameters in the Process's UserData? */
            if ( !("".equals(mqt.getUserData())) )
            {
               xmlFilename = mqt.getUserData();
               MQFM.logger.info("UserData="+xmlFilename);
            }

            /* Are there any parameters in the Queue's TriggerData? */
            if ( !("".equals(mqt.getTriggerData())) )
            {
               xmlFilename = mqt.getTriggerData();
               MQFM.logger.info("TriggerData="+xmlFilename);
            }
            flag = true;
         }
         catch (Exception e)
         {
            MQFM.logger.fatal("Exception: "+ e.getLocalizedMessage());
            flag = false;
         }
      }
      else
         flag = false;

      return flag;
   }

   /**
    * Lookup the MQFM Windows service's name in the property file
    * for the MQFM XML file to be used during this innovation.
    * @return xmlFilename
    */
   private String handleServiceStartup()
   {
      String xmlFilename = MQFM_XML_FILENAME; // set default value
      Properties props = new Properties();

      try
      {
         File ff = new File(Constants.SERVICE_FILE_NAME);
         ff.getAbsolutePath();
         MQFM.logger.info(Constants.SERVICE_FILE_NAME+" absolute path is '" + ff.getAbsolutePath() + "'");
         
         FileInputStream in = new FileInputStream(ff);
         props.load(in);
         in.close();

         String serviceName = System.getProperty("exe4j.launchName");
         if (serviceName != null)
         {
            xmlFilename = props.getProperty(serviceName, MQFM_XML_FILENAME);
            MQFM.logger.info("Found '" + xmlFilename + "' for service name '" + serviceName+"'");
         }
         else
         {
            xmlFilename = MQFM_XML_FILENAME; // set default value
            MQFM.logger.info("Using default value of '" + xmlFilename + "'");
         }
      }
      catch (FileNotFoundException e)
      {
         MQFM.logger.fatal("FileNotFoundException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
         xmlFilename = MQFM_XML_FILENAME; // set default value
         MQFM.logger.info("Using default value of '" + xmlFilename + "'");
      }
      catch (Exception e)
      {
         MQFM.logger.fatal("Exception: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());
         xmlFilename = MQFM_XML_FILENAME; // set default value
         MQFM.logger.info("Using default value of '" + xmlFilename + "'");
      }

      return xmlFilename;
   }

   /**
    * Display how to use this program to the end-user.
    */
   private void usage()
   {
      MQFM.logger.fatal("Usage : " + WHO_AM_I + " mqfm_filename.xml");
      //
      System.err.println();
      System.err.println("Usage : " + WHO_AM_I + " mqfm_filename.xml");
      System.err.println();
   }

   /**
    * This class will handle the shutdown.
    *
    */
   private static class ShutdownThread extends Thread
   {
      /**
       * The constructor
       */
      public ShutdownThread()
      {
      }

      /**
       * Do the work.
       */
      public void run()
      {
         MQFM.logger.debug("MQFM: ShutdownThread.");
         
         try
         {
            if(processLock)
            {
               MQFM.logger.info("MQFM: Shutdown - handle process file lock.");
               IO.deleteFile(processFileLock);
            }
         }
         catch (MQFMException e)
         {
            MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         }
      }
   }
}
