package biz.capitalware.mqfm.action.objects;

import java.util.ArrayList;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.handlers.Tokens;
import biz.capitalware.mqfm.mq.objects.QueueObject;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.Cryptography;

/**
 * This class holds the values for the Receive action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ReceiveObject extends FMFileMQObject
{
   public QueueObject  backOutQName = null;
   public String           defaultDirectory = Constants.HOME_PATH;
   public String           defaultFilename = null;
   public String           defaultJobXmlFile = null;
   public char             processType = Constants.ONE_PASS;
   public boolean          getWithConvert = false;
   public boolean          overrideDirectoryName = false;
   public boolean          overrideFileName = false;
   public boolean          overrideJobName = false;
   public boolean          createDir = false;
   public boolean          append = false;
   public Cryptography     crypto = null;

   /**
    * The constructor
    */
   public ReceiveObject()
   {
      super();
   }

   /**
    * Process tokens in Action Object's variables
    */
   public void performTokenization()
   {
      if (backOutQName != null)
      {
         if (backOutQName.qName != null)
            backOutQName.qName = Tokens.replace(backOutQName.qName, MQFM.userProperties);
         
         if (backOutQName.qMgrName != null)
            backOutQName.qMgrName = Tokens.replace(backOutQName.qMgrName, MQFM.userProperties);
      }
   }

   /**
    * Check the user specified parameters.
    */
   public void checkObject()
   {
      if ( overrideJobName && (defaultJobXmlFile == null))
         throw new IllegalArgumentException("Error: You must specify a Default / XmlJobFile to use the 'override' attribute.");

      if ( overrideFileName && (defaultFilename == null))
         throw new IllegalArgumentException("Error: You must specify a Default / FileName to use the 'override' attribute.");

      if (crypto != null)
      {
         if ( (crypto.keySize != 128) && (crypto.keySize != 192) && (crypto.keySize != 256) )
            throw new IllegalArgumentException("Error: You must specify a keysize attribute that is 128 or 192 or 256 for the Send action.");

         if ("".equals(crypto.passPhrase))
            throw new IllegalArgumentException("Error: PassPhrase attribute cannot be blank for the Send action.");
      }
   }

   /**
    * Print out the user supplied parameters.
    */
   public void printValues()
   {
      MQFM.logger.info("     GetWithConvert       = " + getWithConvert);
      MQFM.logger.info("     BackOutQName         = " + (backOutQName==null?"":backOutQName));
      MQFM.logger.info("     Default JobFile      = " + (defaultJobXmlFile==null?"":defaultJobXmlFile));
      MQFM.logger.info("     Default Directory Path= " + defaultDirectory);
      MQFM.logger.info("     Default Filename     = " + (defaultFilename==null?"":defaultFilename));
      MQFM.logger.info("     Override JobXmlFile  = " + overrideJobName);
      MQFM.logger.info("     Override Directory   = " + overrideDirectoryName);
      MQFM.logger.info("     Override File Name   = " + overrideFileName);
      MQFM.logger.info("     Append               = " + append);
      MQFM.logger.info("     Create Directory     = " + createDir);

      String type = "Invalid";
      if (processType == Constants.ONE_PASS)
         type = "OnePass";
      else if (processType == Constants.RUN_UNTIL_EMPTY)
         type = "RunUntilEmpty";
      else if (processType == Constants.TRIGGERED)
         type = "Triggered";
      else if (processType == Constants.DAEMON)
         type = "Daemon";

      MQFM.logger.info("     Processing Type      = " + type);
      
      if (crypto != null)
      {
         MQFM.logger.info("     KeySize            = " + crypto.keySize);
         MQFM.logger.info("     PassPhrase         = " + crypto.passPhrase );
      }
   }
   
   /**
    * Convert the object to a string.
    * @return s
    */
   public ArrayList<String> toStringArray()
   {
      ArrayList<String> al = new ArrayList<String>();

      al.add("<"+Constants.RECEIVE_ELEMENT+" "+Constants.GETWITHCONVERT_ATTRIBUTE+"=\""+getWithConvert+"\" "+Constants.RUN_ATTRIBUTE+"=\""+processType+"\" "+Constants.APPEND_ATTRIBUTE+"=\""+append+"\""+Constants.CREATE_DIR_ATTRIBUTE+"=\""+createDir+"\">"+LINE_SEPARATOR);

      /*
       * MQ info
       */
      ArrayList<String> mq = createMQTag();
      for (int i=0; i < mq.size(); i++)
         al.add(mq.get(i));
      
      /*
       * Default info
       */
      if ( (defaultJobXmlFile != null) || (defaultDirectory != null) || (defaultFilename != null) )
      {
         al.add("  <"+Constants.DEFAULT_ELEMENT+">"+LINE_SEPARATOR);

         if (defaultJobXmlFile != null)
            al.add("    <"+Constants.EXECUTE_ELEMENT+">"+defaultJobXmlFile+"</"+Constants.EXECUTE_ELEMENT+">"+LINE_SEPARATOR);

         if (defaultDirectory != null)
            al.add("    <"+Constants.DIRECTORY_ELEMENT+">"+defaultDirectory+"</"+Constants.DIRECTORY_ELEMENT+">"+LINE_SEPARATOR);

         if (defaultFilename != null)
            al.add("    <"+Constants.FILENAME_ELEMENT+">"+defaultFilename+"</"+Constants.FILENAME_ELEMENT+">"+LINE_SEPARATOR);
         
         al.add("  </"+Constants.DEFAULT_ELEMENT+">"+LINE_SEPARATOR);
      }

      al.add("</"+Constants.RECEIVE_ELEMENT+">"+LINE_SEPARATOR);
      
      return al;
   }
}
