package biz.capitalware.mqfm.action.objects;

import java.io.File;
import java.util.ArrayList;
import java.util.StringTokenizer;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.handlers.StringUtilities;
import biz.capitalware.mqfm.handlers.Tokens;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.StartEnd;

/**
 * This class holds the values for the ReplaceText action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ReplaceTextObject extends FMFileObject
{
   public String     srcFilename = null;
   public String     toFilename = null;
   public String     findText = null;
   public String     replaceText = null;
   public StartEnd   row = new StartEnd();
   public StartEnd   column = new StartEnd();
   public boolean    createDir = false;

   /**
    * The constructor
    */
   public ReplaceTextObject()
   {
      super();
   }

   /**
    * Process tokens in Action Object's variables
    */
   public void performTokenization()
   {
      if (srcFilename != null)
         srcFilename   = Tokens.replace(srcFilename, MQFM.userProperties);
      if (toFilename != null)
         toFilename = Tokens.replace(toFilename, MQFM.userProperties);
   }

   /**
    * Verify the values in the object.
    */
   public void checkObject()
   {
      String targetDir = null;
      if (toFilename.indexOf(File.separator) != -1)
         targetDir = toFilename.substring(0,toFilename.lastIndexOf(File.separator));

      /* Do we need to create the directory? */
      if (targetDir != null)
      {
         if (createDir)
         {
            if (!(new File(targetDir).exists()))
            {
               if ((new File(targetDir)).mkdirs())
                  MQFM.logger.info("Successfully created directory: " + targetDir);
               else
                  throw new IllegalArgumentException("Error: Cannot create directory: " + targetDir);
            }
         }
         else if (!(new File(targetDir).exists()))
            throw new IllegalArgumentException("Target directory '"+targetDir+"' does not exit.");
      }
      
      if (srcFilename == null)
         throw new IllegalArgumentException("Error: You must specify a filename for the ReplaceText action.");

      if (toFilename == null)
         throw new IllegalArgumentException("Error: You must specify a toFilename for the ReplaceText action.");

      if ( (findText == null) || ("".equals(findText)) )
         throw new IllegalArgumentException("Error: You must specify a find for the ReplaceText action.");

      if ( (replaceText == null) || ("".equals(replaceText)) )
         throw new IllegalArgumentException("Error: You must specify a replace for the ReplaceText action.");

      if (row.start < 1)
         throw new IllegalArgumentException("Error: You must specify a row start value that is greater then zero for the ReplaceText action.");

      if ( (row.end < -1) || (row.end == 0) )
         throw new IllegalArgumentException("Error: You must specify a row end value that is greater then zero for the ReplaceText action.");

      if (column.start < 1)
         throw new IllegalArgumentException("Error: You must specify a column start value that is greater then zero for the ReplaceText action.");

      if ( (column.end < -1) || (column.end == 0) )
         throw new IllegalArgumentException("Error: You must specify a column end value that is greater then zero for the ReplaceText action.");
   }

   /**
    * Output the info
    * @param logger
    */
   public void printValues()
   {
      MQFM.logger.info("     FileName             = " + srcFilename);
      MQFM.logger.info("     ToFileName           = " + toFilename);
      MQFM.logger.info("     Find                 = " + findText);
      MQFM.logger.info("     Replace              = " + replaceText);
      MQFM.logger.info("     Rows                 = " + row.start+":"+(row.end == -1?"*":row.end));
      MQFM.logger.info("     Columns              = " + column.start+":"+(column.end == -1?"*":column.end));
      MQFM.logger.info("     Create Directory     = " + createDir);
   }
   
   /**
    * Split and convert value
    * @param value
    * @param start
    * @param end
    */
   public void convertRTValue(String value, StartEnd se)
   {
      if ( ("*".equals(value)) || ("1:*".equals(value)) )
      {
         se.start = 1;
         se.end = -1;
      }
      else
      {
         StringTokenizer st = new StringTokenizer(value,":");
         
         if (st.hasMoreTokens())
         {
            se.start = StringUtilities.stringToInt(st.nextToken(), 1);
            
            if (st.hasMoreTokens())
               se.end = StringUtilities.stringToInt(st.nextToken(), -1);
            else
               se.end = -1;
         }
         else
         {
            se.start = 1;
            se.end = -1;
         }
      }
   }
   
   /**
    * Convert the object to a string.
    * @return s
    */
   public ArrayList<String> toStringArray()
   {
      ArrayList<String> al = new ArrayList<String>();
      String file;
      String r = row.start+":"+(row.end == -1?"*":row.end);
      String c = column.start+":"+(column.end == -1?"*":column.end);

      if (srcFilename.indexOf(File.separator) != -1 )
      {
         int x = srcFilename.lastIndexOf(File.separator);

         String dir   = srcFilename.substring(0, x);
         String fname = srcFilename.substring(x+1);

         file = "<"+Constants.FILE_ELEMENT+" "+Constants.DIR_ATTRIBUTE+"=\""+dir+"\">"+fname+"</"+Constants.FILE_ELEMENT+">";
      }
      else
         file = "<"+Constants.FILE_ELEMENT+">"+srcFilename+"</"+Constants.FILE_ELEMENT+">";
      
      al.add("<"+Constants.REPLACE_TEXT_ELEMENT+" "+Constants.FILE_ATTRIBUTE+"=\""+toFilename+" "+Constants.ROWS_ATTRIBUTE+"=\""+r+"\" "+Constants.COLUMNS_ATTRIBUTE+"=\""+c+"\">"+LINE_SEPARATOR);
      al.add("  <"+Constants.FIND_ELEMENT+">"+findText+"</"+Constants.FIND_ELEMENT+">"+LINE_SEPARATOR);
      al.add("  <"+Constants.REPLACE_ELEMENT+">"+replaceText+"</"+Constants.REPLACE_ELEMENT+">"+LINE_SEPARATOR);
      al.add("  "+file+LINE_SEPARATOR);
      al.add("</"+Constants.REPLACE_TEXT_ELEMENT+">"+LINE_SEPARATOR);
      
      return al;
   }
}
