package biz.capitalware.mqfm.action.objects;

import java.io.File;
import java.util.ArrayList;

import com.ibm.mq.MQC;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.Cryptography;

/**
 * This class holds the values for the Watch action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class WatchObject extends FMFileMQObject
{
   public String        xmlFile = null;
   public char          format = ' ';
   public String        jobXmlFile = null;
   public String        remoteJobXmlFile = null;
   public String        remoteDirectory = null;
   public String        remoteFilename = null;
   public String        archiveDirectory = Constants.ARCHIVE_PATH;
   public String        archiveFileName = null;
   public int           priority = MQC.MQPRI_PRIORITY_AS_Q_DEF;
   public boolean       persistence = true;
   public boolean       deleteFile = false;
   public boolean       archiveFile = false;
   public boolean       archiveCreateDir = false;
   public Cryptography  crypto = null;

   /**
    * The constructor
    */
   public WatchObject()
   {
      super();
   }

   /**
    * Process tokens in Action Object's variables
    */
   public void performTokenization()
   {
      
   }

   /**
    * Check the user specified parameters.
    */
   public void checkObject()
   {
      if (remoteDirectory != null)
      {
         if ( remoteDirectory.startsWith("jobs"))
            throw new IllegalArgumentException("Security risk. Remote directory cannot be the \"jobs\" directory.");
      }

      if (xmlFile != null)
      {
         File wfn = new File(Constants.WATCH_PATH + xmlFile);
         if (!(wfn.exists()))
               throw new IllegalArgumentException("Watch file name '"+xmlFile+"' does not exit in directory '"+ Constants.WATCH_PATH + "'.");
      }
      else
         throw new IllegalArgumentException("Missing Watch file name.");

      if ( deleteFile &&  archiveFile)
         throw new IllegalArgumentException("Error: User cannot specify both delete and archive together.");

      if (jobXmlFile != null)
      {
         File ljn = new File(Constants.JOBS_PATH + jobXmlFile);
         if (!(ljn.exists()))
               throw new IllegalArgumentException("JobFile '"+jobXmlFile+"' does not exit.");
      }

      if (remoteDirectory == null)
         remoteDirectory = Constants.HOME_PATH;  /* set default values */

      if (crypto != null)
      {
         if ( (crypto.keySize != 128) && (crypto.keySize != 192) && (crypto.keySize != 256) )
            throw new IllegalArgumentException("Error: You must specify a keysize attribute that is 128 or 192 or 256 for the Send action.");

         if ("".equals(crypto.passPhrase))
            throw new IllegalArgumentException("Error: PassPhrase attribute cannot be blank for the Send action.");
      }
   }

   /**
    * Output the info related to File2Msg
    * @param logger
    */
   public void printValues()
   {
      MQFM.logger.info("     ---- Global Watch values ---- ");
      MQFM.logger.info("     Watch Path & Filename= " + xmlFile);
      MQFM.logger.info("     JobFile              = " + (jobXmlFile == null ? "" : jobXmlFile));
      MQFM.logger.info("     Delete File          = " + (deleteFile?"Yes":"No"));
      MQFM.logger.info("     Archive File         = " + (archiveFile?"Yes":"No"));
      MQFM.logger.info("     Archive Create Dir   = " + (archiveCreateDir?"Yes":"No"));
      MQFM.logger.info("     Archive Directory    = " + archiveDirectory);
      MQFM.logger.info("     Archive Filename     = " + (archiveFileName == null ? "" : archiveFileName));
      MQFM.logger.info("     Remote JobXmlFile    = " + (remoteJobXmlFile == null ? "" : remoteJobXmlFile));
      MQFM.logger.info("     Remote Directory     = " + remoteDirectory);
      MQFM.logger.info("     Remote Filename      = " + (remoteFilename == null ? "" : remoteFilename));
      MQFM.logger.info("     Format               = " + format);
      MQFM.logger.info("     Priority             = " + (priority == MQC.MQPRI_PRIORITY_AS_Q_DEF? "MQPRI_PRIORITY_AS_Q_DEF" : priority));
      MQFM.logger.info("     Persistence          = " + (persistence? "Yes":"No"));
      if (crypto != null)
      {
         MQFM.logger.info("     KeySize              = " + crypto.keySize);
         MQFM.logger.info("     PassPhrase           = " + crypto.passPhrase );
      }
   }
   
   /**
    * Convert the object to a string.
    * @return s
    */
   public ArrayList<String> toStringArray()
   {
      ArrayList<String> al = new ArrayList<String>();
      
      if (crypto != null)
         al.add("<"+Constants.WATCH_ELEMENT+" "+Constants.DELETE_ATTRIBUTE+"=\""+(deleteFile?"Y":"N")+"\" "+Constants.FORMAT_ATTRIBUTE+"=\""+format+"\" "+Constants.PRIORITY_ATTRIBUTE+"=\""+priority+"\" "+Constants.PERSISTENCE_ATTRIBUTE+"=\""+(persistence? "Y":"N")+"\" "+Constants.KEY_SIZE_ATTRIBUTE+"=\""+crypto.keySize+"\" "+Constants.PASS_PHRASE_ATTRIBUTE+"=\""+crypto.passPhrase+"\">"+LINE_SEPARATOR);
      else
         al.add("<"+Constants.WATCH_ELEMENT+" "+Constants.DELETE_ATTRIBUTE+"=\""+(deleteFile?"Y":"N")+"\" "+Constants.FORMAT_ATTRIBUTE+"=\""+format+"\" "+Constants.PRIORITY_ATTRIBUTE+"=\""+priority+"\" "+Constants.PERSISTENCE_ATTRIBUTE+"=\""+(persistence? "Y":"N")+"\">"+LINE_SEPARATOR);

      /*
       * MQ info
       */
      ArrayList<String> mq = createMQTag();
      for (int i=0; i < mq.size(); i++)
         al.add(mq.get(i));

      /*
       * Remote info
       */
      ArrayList<String> rt = createRemoteTag(remoteJobXmlFile, remoteDirectory, remoteFilename);
      for (int i=0; i < rt.size(); i++)
         al.add(rt.get(i));
      
      if (jobXmlFile != null)
         al.add("  <"+Constants.EXECUTE_ELEMENT+" "+Constants.XML_FILE_ATTRIBUTE+"=\""+jobXmlFile+"\" />"+LINE_SEPARATOR);
      
      if (archiveFile)
         al.add("  <"+Constants.ARCHIVE_ELEMENT+" "+Constants.CREATE_DIR_ATTRIBUTE+"=\""+(archiveCreateDir?'Y':'N')+"\" "+" "+Constants.TO_DIR_ATTRIBUTE+"=\""+archiveDirectory+"\" "+Constants.TO_FILE_ATTRIBUTE+"=\""+archiveFileName+"\" />"+LINE_SEPARATOR);
      
      al.add("</"+Constants.WATCH_ELEMENT+">"+LINE_SEPARATOR);
      
      return al;
   }
}
