package biz.capitalware.mqfm.actions;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileLock;
import java.util.zip.Deflater;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.action.objects.ZipObject;
import biz.capitalware.mqfm.handlers.DirectoryRecurser;
import biz.capitalware.mqfm.handlers.IO;
import biz.capitalware.mqfm.handlers.StringUtilities;
import biz.capitalware.mqfm.handlers.Tokens;

/**
 * This class will handle the Zip action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ZipAction extends FMAction
{
   private ZipObject   zo;
   private String      osName;

   /**
    * The constructor
    * @param zo
    * @param logger
    */
   public ZipAction(ZipObject zo)
   {
      super(zo.actionName);
      this.zo = zo;
      osName = System.getProperty("os.name");
   }

   /**
    * Perform the zip work.
    */
   public int doWork()
   {
      int exitCode = 0;
      zo.printValues();

      zo.zipFilename = Tokens.replace(zo.actionName, zo.zipFilename, zo.srcFilename, 1, MQFM.userProperties);

      File source = new File(zo.srcFilename);

      try
      {
         if (source.isDirectory())
            zipDirectoryOfFiles();
         else if (zo.srcFilename.indexOf('*') != -1)
            zipManyFiles(zo.srcFilename);
         else
            zipOneFile();

         zo.fileCount++;

         MQFM.logger.info("Successfully created archive '" + zo.zipFilename+"' from '"+zo.srcFilename+"'");
      }
      catch (IOException e)
      {
         MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         MQFM.globalErrorText = e.getLocalizedMessage();
         
         if (zo.onErrorFail)
            exitCode = 8;
      }

      return exitCode;
   }

   /**
    * Perform creating archive from a directory..
    * @throws IOException
    * @throws FileNotFoundException
    */
   private void zipDirectoryOfFiles() throws IOException, FileNotFoundException
   {
      DirectoryRecurser dr = null;
      FileOutputStream fileout = null;
      FileInputStream fin = null;
      ZipOutputStream zipout = null;
      File f = null;
      ZipEntry ze = null;

      try
      {
         dr = new DirectoryRecurser(new File(zo.srcFilename));
         fileout = new FileOutputStream(zo.zipFilename);
         zipout = new ZipOutputStream(fileout);
         zipout.setLevel(Deflater.BEST_COMPRESSION);
         while (dr.hasMoreElements())
         {
            f = dr.nextElement();
            if (f.isFile())
            {
               ze = new ZipEntry(handleFileName(f.getPath()));
               fin = new FileInputStream(f);
               
               if (isFileLockable(fin, f.getPath()))
               {
                  zipout.putNextEntry(ze);
                  IO.streamCopier(fin, zipout);
               }
               
               fin.close();
               zipout.closeEntry();
            }
         }
         zipout.close();
         fileout.close();
      }
      finally
      {
         try { zipout.close(); } catch (Exception e) {;}
         try { fin.close(); } catch (Exception e) {;}
         try { fileout.close(); } catch (Exception e) {;}
      }
   }

   /**
    * Perform creating archive from many files using a file filter.
    * @throws IOException
    * @throws FileNotFoundException
    */
   private void zipManyFiles(String zipSource) throws IOException, FileNotFoundException
   {
      FileOutputStream fileout = null;
      FileInputStream fin = null;
      ZipOutputStream zipout = null;
      ZipEntry ze = null;

      File[] fileList = IO.getFilteredFileList(zipSource);

      try
      {
         fileout = new FileOutputStream(zo.zipFilename);
         zipout = new ZipOutputStream(fileout);
         zipout.setLevel(Deflater.BEST_COMPRESSION);

         for (int i=0; i < fileList.length; i++)
         {
            File f = new File(fileList[i].toString());
            if (f.isFile())
            {
               ze = new ZipEntry(handleFileName(f.getPath()));
               fin = new FileInputStream(f);
               
               if (isFileLockable(fin, f.getPath()))
               {
                  zipout.putNextEntry(ze);
                  IO.streamCopier(fin, zipout);
               }

               fin.close();
               zipout.closeEntry();
            }
         }

         zipout.close();
         fileout.close();
      }
      finally
      {
         try { zipout.close(); } catch (Exception e) {;}
         try { fin.close(); } catch (Exception e) {;}
         try { fileout.close(); } catch (Exception e) {;}
      }
   }

   /**
    * Perform archiving for a file.
    * @throws IOException
    * @throws FileNotFoundException
    */
   private void zipOneFile() throws IOException, FileNotFoundException
   {
      File source = new File(zo.srcFilename);

      FileOutputStream fileout = null;
      FileInputStream fin = null;
      ZipOutputStream zipout = null;
      ZipEntry ze = null;

      try
      {
         fileout = new FileOutputStream(zo.zipFilename);
         zipout = new ZipOutputStream(fileout);
         zipout.setLevel(Deflater.BEST_COMPRESSION);
         ze = new ZipEntry(handleFileName(source.getPath()));
         fin = new FileInputStream(source);
         
         if (isFileLockable(fin, source.getPath()))
         {
            zipout.putNextEntry(ze);
            IO.streamCopier(fin, zipout);
         }

         fin.close();
         zipout.closeEntry();
         zipout.close();
         fileout.close();
      }
      finally
      {
         try { zipout.close(); } catch (Exception e) {;}
         try { fin.close(); } catch (Exception e) {;}
         try { fileout.close(); } catch (Exception e) {;}
      }
   }
   
   /**
    * Handle Windows drive (C:) and absolute paths.    
    * @param inFileName
    * @return filename
    */
   private String handleFileName(String inFileName)
   {
      String outFileName = inFileName;
      if (osName != null)
      {
         // Strip off drive letters!
         if (osName.toLowerCase().startsWith("windows"))
         {
            if (outFileName.length() > 2)
            {
               char ch1 = outFileName.charAt(0);
               char ch2 = outFileName.charAt(1);
               if (ch2 == ':' && (ch1 >= 'a' && ch1 <= 'z' || ch1 >= 'A' && ch1 <= 'Z'))
               {
                  outFileName = StringUtilities.replaceAll(outFileName.substring(2), "\\", "/");
               }
            }
         }
      }

      // No absolute pathnames
      while (outFileName.startsWith("/"))
      {
         outFileName = outFileName.substring(1);
      }

      return outFileName;
   }   

   /**
    * Can we lock the file?  Or does another process have it?
    * @param fin
    * @param fileName
    * @return flag
    * @throws IOException
    */
   private boolean isFileLockable(FileInputStream fin, String fileName) throws IOException
   {
      boolean flag = false;
      FileLock fl = null;

      try
      {
         fl = fin.getChannel().tryLock(0L, Long.MAX_VALUE, true);
         if(fl != null)
           flag = true;
         else
            MQFM.logger.warn("Warning: File is locked by another process - skipping file: "+fileName);
      }
      finally
      {
         if (fl != null)
            fl.release();
      }

      return flag;
   }
}
