package biz.capitalware.mqfm.mq;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import org.jdom.JDOMException;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.action.objects.ReceiveObject;
import biz.capitalware.mqfm.handlers.AES;
import biz.capitalware.mqfm.handlers.IO;
import biz.capitalware.mqfm.handlers.RemoteValuesInMsg;
import biz.capitalware.mqfm.handlers.RunJob;
import biz.capitalware.mqfm.handlers.JZip;
import biz.capitalware.mqfm.mq.objects.MQObject;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.MQFMException;

import com.ibm.mq.*;

/**
 * This class will handle all MQ interactions for getting message(s) and writing 
 * it to a file.
 *
 * @author Roger Lacroix
 * @version 1.4
 * @license Apache 2 License
 */
public class MQGetMsg extends MyMQ
{
   private static boolean  loopAgain = true;
   private        int      msgCounter = 0;

   /**
    * The constructor
    */
   public MQGetMsg(MQObject qm)
   {
   	super(qm);
   }

   /**
    * Exit loop processing
    */
   public void stopIt()
   {
      loopAgain = false;
   }
   
   /**
    * Retrieve a message from a queue and write it to a file.
    * @param ro
    * @throws MQFMException 
    */
   public int retrieveMsg(ReceiveObject ro) throws MQFMException
   {
      int exitCode = 0;
      MQQueue inQ = null;
      MQQueue boQ = null;
      MQGetMessageOptions gmo = new MQGetMessageOptions();
      gmo.matchOptions = MQC.MQMO_NONE;
      gmo.options = MQC.MQGMO_FAIL_IF_QUIESCING + MQC.MQGMO_SYNCPOINT + MQC.MQGMO_LOGICAL_ORDER;

      if (ro.getWithConvert)
         gmo.options += MQC.MQGMO_CONVERT;

      if (ro.processType == Constants.DAEMON)
      {
         gmo.options += MQC.MQGMO_WAIT;
         gmo.waitInterval = 60000; // wait up to 60 seconds
      }

      inQ = openInQ(mqo.qObject);

      if (ro.backOutQName != null)
         boQ = openOutQ(ro.backOutQName);

      if ( (inQ == null) || 
           ( (ro.backOutQName != null) && (boQ == null) ) )
         loopAgain = false;

      MQMessage receiveMsg = new MQMessage();
      receiveMsg.messageId     = MQC.MQMI_NONE;
      receiveMsg.correlationId = MQC.MQCI_NONE;
      receiveMsg.groupId       = MQC.MQGI_NONE;
      
      try
      {
         while (loopAgain)
         {
            try
            {
               inQ.get(receiveMsg, gmo);
               ro.msgCount++;
   
               if (receiveMsg.feedback == MQC.MQFB_QUIT)
               {
                  loopAgain = false;
                  break;
               }
               
               if ((receiveMsg.messageFlags & MQC.MQMF_MSG_IN_GROUP) == MQC.MQMF_MSG_IN_GROUP)
               {
                  handleGroupOfMsgs(ro, receiveMsg, inQ, boQ);
               }
               else
               {
                  handleOutputtingSingleFile(ro, receiveMsg, boQ);
               }
               
               receiveMsg.clearMessage();
               receiveMsg.messageId     = MQC.MQMI_NONE;
               receiveMsg.correlationId = MQC.MQCI_NONE;
               receiveMsg.groupId       = MQC.MQGI_NONE;
            }
            catch (JDOMException je)
            {
               MQFM.logger.fatal("JDOMException: " + je.getLocalizedMessage());
               if (je.getCause() != null)
                  MQFM.logger.fatal(je.getCause().toString());
               MQFM.globalErrorText = je.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (MQException e)
            {
               if ( (ro.processType == Constants.TRIGGERED)         &&
                    (e.completionCode == MQException.MQCC_FAILED)     &&
                    (e.reasonCode == MQException.MQRC_NO_MSG_AVAILABLE) )
               {
                  MQFM.logger.info("Triggering loop : no more messages.");
                  loopAgain = false;
               }
               else if ( (ro.processType == Constants.ONE_PASS)         &&
                         (e.completionCode == MQException.MQCC_FAILED)     &&
                         (e.reasonCode == MQException.MQRC_NO_MSG_AVAILABLE) )
                {
                   MQFM.logger.info("One Pass loop : no more messages.");
                   loopAgain = false;
                }
               else if ( (ro.processType == Constants.RUN_UNTIL_EMPTY)   &&
                         (e.completionCode == MQException.MQCC_FAILED)     &&
                         (e.reasonCode == MQException.MQRC_NO_MSG_AVAILABLE) )
                {
                   MQFM.logger.info("RunUntilEmpty loop : no more messages.");
                   loopAgain = false;
                }
               else if ( (ro.processType == Constants.DAEMON)            &&
                         (e.completionCode == MQException.MQCC_FAILED)     &&
                         (e.reasonCode == MQException.MQRC_NO_MSG_AVAILABLE) )
               {
                  MQFM.logger.info("Daemon loop : no message found.");
               }
               else
               {
                  MQFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
                  MQFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
   
                  MQFM.globalErrorText = e.getLocalizedMessage();
                  loopAgain = handleProblem(ro, receiveMsg, boQ);
                  if (!loopAgain)
                     exitCode = 99;
               }
            }
            catch (MQFMException e)
            {
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (IOException e)
            {
               MQFM.logger.fatal("IOException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (InvalidKeyException e)
            {
               MQFM.logger.fatal("InvalidKeyException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (NoSuchAlgorithmException e)
            {
               MQFM.logger.fatal("NoSuchAlgorithmException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (NoSuchPaddingException e)
            {
               MQFM.logger.fatal("NoSuchPaddingException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (InvalidAlgorithmParameterException e)
            {
               MQFM.logger.fatal("InvalidAlgorithmParameterException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (IllegalBlockSizeException e)
            {
               MQFM.logger.fatal("IllegalBlockSizeException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
            catch (BadPaddingException e)
            {
               MQFM.logger.fatal("BadPaddingException: " + e.getMessage());
               MQFM.globalErrorText = e.getLocalizedMessage();
   
               loopAgain = handleProblem(ro, receiveMsg, boQ);
               if (!loopAgain)
                  exitCode = 99;
            }
   
            if (ro.processType == Constants.ONE_PASS)
               loopAgain = false;
         }
      }
      finally
      {
         closeQ(inQ);
         closeQ(boQ);
      }

      MQFM.logger.info("Receive Action processed "+msgCounter+" messages.");

      return exitCode;
   }

   /**
    * Retrieve all messages in a group and write them to a file.
    * @param ro
    * @param receiveMsg
    * @param msgCounter
    * @param inQ
    * @param boQ
    * @throws IOException 
    * @throws JDOMException 
    * @throws MQFMException 
    * 
    */
   private void handleGroupOfMsgs(ReceiveObject ro, MQMessage receiveMsg, MQQueue inQ, MQQueue boQ) throws JDOMException, IOException, MQFMException
   {
      boolean loopMsgInGroup = true;
      boolean processIt = true;
      int msgNum = 1;
      long fileSize = 0;
      byte[] b;
      RemoteValuesInMsg rv = new RemoteValuesInMsg();
      rv.processValues(ro, receiveMsg, msgCounter);
      
      MQGetMessageOptions gmo = new MQGetMessageOptions();
      gmo.matchOptions = MQC.MQMO_MATCH_GROUP_ID;
      gmo.waitInterval = 5000;  // 5 Seconds
      gmo.options = MQC.MQGMO_FAIL_IF_QUIESCING + MQC.MQGMO_SYNCPOINT + MQC.MQGMO_WAIT + MQC.MQGMO_LOGICAL_ORDER;

      if (ro.getWithConvert)
         gmo.options += MQC.MQGMO_CONVERT;

      /* Create the directory if it does not exist? */
      if (ro.createDir)
      {
         if (!(new File(rv.getDirectory())).exists())
         {
            if ((new File(rv.getDirectory())).mkdirs()) 
               MQFM.logger.info("Successfully created directory: " + rv.getDirectory());
            else
               throw new IOException("handleGroupOfMsgs(): Cannot create directory: " + rv.getDirectory());
         }
      }
      
      RandomAccessFile out = new RandomAccessFile(rv.getToPathAndFilename(), "rw");
      FileChannel fileChannel = out.getChannel();
      FileLock fileLock = fileChannel.tryLock(); /* exclusive lock for writing */

      if (fileLock == null)
      {
         MQFM.logger.fatal("handleGroupOfMsgs cannot acquire lock for file: " + rv.getToPathAndFilename());
         MQFM.globalErrorText = "handleGroupOfMsgs cannot acquire lock for file: " + rv.getToPathAndFilename();
         throw new MQFMException("handleGroupOfMsgs cannot acquire lock for file: " + rv.getToPathAndFilename());
      }
      
      if ( ((receiveMsg.feedback & Constants.MQFM_ENCRYPT_MSG) == Constants.MQFM_ENCRYPT_MSG) &&
            (ro.crypto == null) )
      {
         MQFM.logger.error("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
         if (handleProblem(ro, receiveMsg, boQ))
            processIt = false;
         else
            throw new MQFMException("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
      }
      
      while (loopMsgInGroup)
      {
         try
         {
            if (processIt)
            {
               if ( (msgNum == 1) &&
                    (receiveMsg.feedback & Constants.MQFM_XML_EMBEDDED_MSG) == Constants.MQFM_XML_EMBEDDED_MSG)
                  b = new byte[receiveMsg.getMessageLength() - rv.getXmlDataLength()];
               else
                  b = new byte[receiveMsg.getMessageLength()];
               
               fileSize += b.length;
               receiveMsg.readFully(b);
               msgCounter++;
               
               if ((receiveMsg.feedback & Constants.MQFM_ZIP_ENCRYPT_MSG) == Constants.MQFM_ZIP_ENCRYPT_MSG)
               {
                  AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
                  b = JZip.decompressData(aes.decrypt(b));
               }
               else if ((receiveMsg.feedback & Constants.MQFM_ENCRYPT_MSG) == Constants.MQFM_ENCRYPT_MSG)
               {
                  AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
                  b = aes.decrypt(b);
               }
               else if ((receiveMsg.feedback & Constants.MQFM_ZIP_MSG) == Constants.MQFM_ZIP_MSG)
                  b = JZip.decompressData(b);
                  
               out.write(b);
               fileSize += b.length;
   
               _qMgr.commit();
   
               if ((receiveMsg.messageFlags & MQC.MQMF_LAST_MSG_IN_GROUP) == MQC.MQMF_LAST_MSG_IN_GROUP)
                  MQFM.logger.info("Retrieved last message # "+msgNum+" and wrote it to '" + rv.getToPathAndFilename() + "'" );
               else
                  MQFM.logger.info("Retrieved message # "+msgNum+" and wrote it to '" + rv.getToPathAndFilename() + "'" );
            }
            /**
             * Setup and get the next message in group.
             */
            processIt = true;
            receiveMsg.clearMessage();
            receiveMsg.messageId     = MQC.MQMI_NONE;
            receiveMsg.correlationId = MQC.MQCI_NONE;
            
            inQ.get(receiveMsg, gmo);
            msgNum++;
            ro.msgCount++;

            if ( ((receiveMsg.feedback & Constants.MQFM_ENCRYPT_MSG) == Constants.MQFM_ENCRYPT_MSG) &&
                  (ro.crypto == null) )
            {
               MQFM.logger.error("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
               if (handleProblem(ro, receiveMsg, boQ))
                  processIt = false;
               else
                  throw new MQFMException("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
            }
         }
         catch (MQException e)
         {
            if (e.reasonCode == MQException.MQRC_NO_MSG_AVAILABLE)
            {
               loopMsgInGroup = false;
            }
            else
            {
               MQFM.logger.fatal("MQException: " + e.getMessage() + " : QName=" + mqo.qObject);
               MQFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);

               MQFM.globalErrorText = e.getLocalizedMessage();
               loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
            }
         }
         catch (IOException e)
         {
            MQFM.logger.fatal("IOException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
         catch (InvalidKeyException e)
         {
            MQFM.logger.fatal("InvalidKeyException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
         catch (NoSuchAlgorithmException e)
         {
            MQFM.logger.fatal("NoSuchAlgorithmException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
         catch (NoSuchPaddingException e)
         {
            MQFM.logger.fatal("NoSuchPaddingException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
         catch (InvalidAlgorithmParameterException e)
         {
            MQFM.logger.fatal("InvalidAlgorithmParameterException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
         catch (IllegalBlockSizeException e)
         {
            MQFM.logger.fatal("IllegalBlockSizeException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
         catch (BadPaddingException e)
         {
            MQFM.logger.fatal("BadPaddingException: " + e.getMessage());
            MQFM.globalErrorText = e.getLocalizedMessage();

            loopMsgInGroup = handleProblem(ro, receiveMsg, boQ);
         }
      }

      if(fileLock != null) 
         fileLock.release();         

      if (fileChannel != null)
         fileChannel.close();

      out.close();
      ro.fileCount++;

      MQFM.logger.info("Retrieved group of messages and they were written to '" + rv.getToPathAndFilename() + "' and filesize is "+fileSize);
   }

   /**
    * Output a single message to a file
    * @param ro
    * @param receiveMsg
    * @param rv
    * @param msgCounter
    * @throws IOException 
    * @throws MQException 
    * @throws JDOMException 
    * @throws InvalidAlgorithmParameterException 
    * @throws NoSuchPaddingException 
    * @throws NoSuchAlgorithmException 
    * @throws InvalidKeyException 
    * @throws BadPaddingException 
    * @throws IllegalBlockSizeException 
    * @throws MQFMException 
    */
   private void handleOutputtingSingleFile(ReceiveObject ro, MQMessage receiveMsg, MQQueue boQ) throws IOException, MQException, JDOMException, InvalidKeyException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException, MQFMException
   {
      boolean processIt = true;
      byte[] b;
      RemoteValuesInMsg rv = new RemoteValuesInMsg();
      rv.processValues(ro, receiveMsg, msgCounter);

      if ( ((receiveMsg.feedback & Constants.MQFM_ENCRYPT_MSG) == Constants.MQFM_ENCRYPT_MSG) &&
           (ro.crypto == null) )
      {
         MQFM.logger.error("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
         if (handleProblem(ro, receiveMsg, boQ))
            processIt = false;
         else
            throw new MQFMException("Message is encrypted but the Receive Action is does not contain KeySize and Passphrase.");
      }
      
      if (processIt)
      {
         /* Create the directory if it does not exist? */
         if (ro.createDir)
         {
            if (!(new File(rv.getDirectory())).exists())
            {
               if ((new File(rv.getDirectory())).mkdirs()) 
                  MQFM.logger.info("Successfully created directory: " + rv.getDirectory());
               else
                  throw new IOException("handleOutputtingSingleFile(): Cannot create directory: " + rv.getDirectory());
            }
         }
         
         if ((receiveMsg.feedback & Constants.MQFM_XML_EMBEDDED_MSG) == Constants.MQFM_XML_EMBEDDED_MSG)
            b = new byte[receiveMsg.getMessageLength() - rv.getXmlDataLength()];
         else
            b = new byte[receiveMsg.getMessageLength()];
         
         receiveMsg.readFully(b);
         
         if ((receiveMsg.feedback & Constants.MQFM_ZIP_ENCRYPT_MSG) == Constants.MQFM_ZIP_ENCRYPT_MSG)
         {
            AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
            IO.writeFile(rv.getToPathAndFilename(), JZip.decompressData(aes.decrypt(b)), ro.append);
         }
         else if ((receiveMsg.feedback & Constants.MQFM_ENCRYPT_MSG) == Constants.MQFM_ENCRYPT_MSG)
         {
            AES aes = new AES(ro.crypto.passPhrase, ro.crypto.keySize);
            IO.writeFile(rv.getToPathAndFilename(), aes.decrypt(b), ro.append);
         }
         else if ((receiveMsg.feedback & Constants.MQFM_ZIP_MSG) == Constants.MQFM_ZIP_MSG)
            IO.writeFile(rv.getToPathAndFilename(), JZip.decompressData(b), ro.append);
         else
            IO.writeFile(rv.getToPathAndFilename(), b, ro.append);
         
         MQFM.logger.info("Retrieved message written to '" + rv.getToPathAndFilename() + "'" );
         msgCounter++;
         ro.fileCount++;
   
         _qMgr.commit();
         
         if (rv.isRunJob())
         {
            RunJob rj = new RunJob();
            rj.launch(rv.getJobXmlFile());
         }
      }
   }

   /**
    *
    * @param ro
    * @param receiveMsg
    * @param boQ
    * @return
    */
   private boolean handleProblem(ReceiveObject ro, MQMessage receiveMsg, MQQueue boQ)
   {
      boolean flag = true;

      /*
       * Do we have a backout queue?
       * If so, attempt to move the message to it.
       */
      if (boQ != null)
      {
         MQFM.logger.info("Attempting to move the message to the backout queue: " + ro.backOutQName);
         MQPutMessageOptions pmo = new MQPutMessageOptions();
         pmo.options = MQC.MQPMO_SET_IDENTITY_CONTEXT + MQC.MQPMO_FAIL_IF_QUIESCING;
         try
         {
            MQMessage sendmsg = receiveMsg;
            // put the message on the queue.
            boQ.put(sendmsg, pmo);
            _qMgr.commit();
            MQFM.logger.info("Successfully moved the message to the backout queue: " + ro.backOutQName);
         }
         catch (MQException e1)
         {
            MQFM.logger.fatal("MQException: " + e1.getMessage() + " : QName=" + mqo.qObject);
            MQFM.logger.fatal("MQCC=" + e1.completionCode + " : MQRC="+e1.reasonCode);
            MQFM.globalErrorText = e1.getLocalizedMessage();

            backout();
            flag = false; 
         }
      }
      else
      {
         backout();
         flag = false; 
      }

      return flag;
   }
}
