package biz.capitalware.mqfm.mq;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

import org.jdom.JDOMException;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.action.objects.SendObject;
import biz.capitalware.mqfm.handlers.AES;
import biz.capitalware.mqfm.handlers.IO;
import biz.capitalware.mqfm.handlers.RunJob;
import biz.capitalware.mqfm.handlers.Tokens;
import biz.capitalware.mqfm.handlers.JZip;
import biz.capitalware.mqfm.mq.objects.MQObject;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.MQFMException;

import com.ibm.mq.*;

/**
 * This class will handle all MQ interactions for putting a file as a message onto a queue.
 *
 * @author Roger Lacroix
 * @version 1.4
 * @license Apache 2 License
 */
public class MQPutFile extends MyMQ
{
   /**
    * The constructor
    */
   public MQPutFile(MQObject qm)
   {
   	super(qm);
   }

   /**
    * Do a MQPUT1: Open queue, put a message from a file to the queue and then close it.
    *
    * @param so
    * @throws MQFMException 
    */
   public int simplePut1(SendObject so) throws MQFMException
   {
      int exitCode = 0;
      MQQueue outQ = null;
      MQDistributionList outDL = null;
      
      if (mqo.disListOfQNames != null)
      {
         MQFM.logger.info("simplePut1: Attempting to open distribution list '"+mqo.disListOfQNames+"'");
         outDL = openOutDL(mqo.disListOfQNames);
      }
      else
      {
         MQFM.logger.info("simplePut1: Attempting to open queue '"+mqo.qObject+"'");
         outQ = openOutQ(mqo.qObject);
      }
      
      try
      {
         if ( (outQ != null) || (outDL != null) )
            putMsg(so, outQ, outDL);
         else
            exitCode = 99;
      }
      catch (JDOMException e)
      {
         MQFM.logger.fatal("JDOMException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (InvalidKeyException e)
      {
         MQFM.logger.fatal("InvalidKeyException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (NoSuchAlgorithmException e)
      {
         MQFM.logger.fatal("NoSuchAlgorithmException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (NoSuchPaddingException e)
      {
         MQFM.logger.fatal("NoSuchPaddingException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (InvalidAlgorithmParameterException e)
      {
         MQFM.logger.fatal("InvalidAlgorithmParameterException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (IllegalBlockSizeException e)
      {
         MQFM.logger.fatal("IllegalBlockSizeException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (BadPaddingException e)
      {
         MQFM.logger.fatal("BadPaddingException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      finally
      {
         if (outQ != null)
            closeQ(outQ);
         if (outDL != null)
            closeDL(outDL);
      }

      return exitCode;
   }

   /**
    * This method will get a directory listing of all files to be sent then call simplePut1 for each file.
    * @param so
    * @throws MQFMException 
    * @throws BadPaddingException 
    * @throws IllegalBlockSizeException 
    * @throws InvalidAlgorithmParameterException 
    * @throws NoSuchPaddingException 
    * @throws NoSuchAlgorithmException 
    * @throws InvalidKeyException 
    */
   public int putManyMsgs(SendObject so) throws MQFMException, InvalidKeyException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException
   {
      int exitCode = 0;
      MQQueue outQ = null;
      MQDistributionList outDL = null;
      
      if (mqo.disListOfQNames != null)
      {
         MQFM.logger.info("putManyMsgs: Attempting to open distribution list '"+mqo.disListOfQNames+"'");
         outDL = openOutDL(mqo.disListOfQNames);
      }
      else
      {
         MQFM.logger.info("putManyMsgs: Attempting to open queue '"+mqo.qObject+"'");
         outQ = openOutQ(mqo.qObject);
      }

      try
      {
         if (outQ != null)
         {
            File[] fileList = IO.getFilteredFileList(so.srcFilename);
   
            for (int i=0; i < fileList.length; i++)
            {
               so.srcFilename = fileList[i].toString();
               so.remoteFilename = fileList[i].getName();
   
               putMsg(so, outQ, outDL);
            }
         }
         else
         {
            exitCode = 99;
         }
      }
      catch (JDOMException e)
      {
         MQFM.logger.fatal("JDOMException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      catch (IOException e)
      {
         MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            MQFM.logger.fatal(e.getCause().toString());;
         MQFM.globalErrorText = e.getLocalizedMessage();
         exitCode = 99;
      }
      finally
      {
         if (outQ != null)
            closeQ(outQ);
         if (outDL != null)
            closeDL(outDL);
      }

      return exitCode;
   }

   /**
    * Determine the file size then select the appropriate method to put the file.
    *
    * @param so
    * @param outQ
    * @throws JDOMException
    * @throws BadPaddingException 
    * @throws IllegalBlockSizeException 
    * @throws InvalidAlgorithmParameterException 
    * @throws NoSuchPaddingException 
    * @throws NoSuchAlgorithmException 
    * @throws InvalidKeyException 
    * @throws MQFMException 
    */
   public void putMsg(SendObject so, MQQueue outQ, MQDistributionList outDL) throws JDOMException, InvalidKeyException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException, MQFMException
   {
      if ( (new File(so.srcFilename)).length() <= Constants.MAX_GROUP_MSG_SIZE)
         putMsgLess4MB(so, outQ, outDL);
      else
         putMsgGreater4MB(so, outQ, outDL);
   }

   /**
    * Do a MQPUT1: Open queue, put a message from a file to the queue and then close it.
    *
    * @param so
    * @param outQ
    * @throws JDOMException
    * @throws InvalidAlgorithmParameterException 
    * @throws NoSuchPaddingException 
    * @throws NoSuchAlgorithmException 
    * @throws InvalidKeyException 
    * @throws BadPaddingException 
    * @throws IllegalBlockSizeException 
    * @throws MQFMException 
    */
   public void putMsgLess4MB(SendObject so, MQQueue outQ, MQDistributionList outDL) throws JDOMException, InvalidKeyException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException, MQFMException
   {
      MQPutMessageOptions pmo = new MQPutMessageOptions();
      pmo.options = MQC.MQPMO_SET_IDENTITY_CONTEXT + MQC.MQPMO_FAIL_IF_QUIESCING;
      MQFM.logger.info("putMsgLess4MB: Attempting to send a file '"+so.srcFilename);

      try
      {
         MQMessage sendMsg = createMQFMMessage(so);
         
         if ( (so.crypto != null) && (so.format == Constants.FORMAT_ZIP) ) 
         {
            AES aes = new AES(so.crypto.passPhrase, so.crypto.keySize);
            sendMsg.write(aes.encrypt(JZip.compressData(IO.readFile(so.srcFilename))));
         }
         else if (so.crypto != null) 
         {
            AES aes = new AES(so.crypto.passPhrase, so.crypto.keySize);
//            byte[] b = IO.readFile(so.srcFilename);
//            byte[] e = aes.encrypt(b);
//            sendMsg.write(e);
            
            sendMsg.write(aes.encrypt(IO.readFile(so.srcFilename)));
         }
         else if (so.format == Constants.FORMAT_ZIP)
            sendMsg.write(JZip.compressData(IO.readFile(so.srcFilename)));
         else
            sendMsg.write(IO.readFile(so.srcFilename));

         if (outDL != null)
         {
            outDL.put(sendMsg, pmo); // put the message to distribution list
            MQFM.logger.info("Put file '" + so.srcFilename + "' as a message to distribution list: " + "'"+mqo.disListOfQNames+"'");
         } 
         else
         {
            outQ.put(sendMsg, pmo); // put the message on the queue.
            MQFM.logger.info("Put file '" + so.srcFilename + "' as a message to queue: " + "'"+mqo.qObject+"'");
         }
         
         so.fileCount++;
         so.msgCount++;

         handlePostCommon(so);
      }
      catch (MQException e)
      {
         MQFM.logger.fatal("MQException: " + e.getLocalizedMessage() + " : QObject: " + mqo.qObject);
         MQFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
         MQFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (IOException ioex)
      {
         MQFM.logger.fatal("IOException: " + ioex.getLocalizedMessage());
         MQFM.globalErrorText = ioex.getLocalizedMessage();
      }
   }

   /**
    * Do a MQPUT1: Open queue, put a message from a file to the queue and then close it.
    *
    * @param so
    * @param outQ
    * @throws JDOMException
    * @throws InvalidAlgorithmParameterException 
    * @throws NoSuchPaddingException 
    * @throws NoSuchAlgorithmException 
    * @throws InvalidKeyException 
    * @throws BadPaddingException 
    * @throws IllegalBlockSizeException 
    * @throws MQFMException 
    */
   public void putMsgGreater4MB(SendObject so, MQQueue outQ, MQDistributionList outDL) throws JDOMException, InvalidKeyException, NoSuchAlgorithmException, NoSuchPaddingException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException, MQFMException
   {
      boolean processingOK = false;
      RandomAccessFile in = null;
      FileChannel fileChannel = null;
      FileLock fileLock = null;
      long fileSize = 0;
      MQPutMessageOptions pmo = new MQPutMessageOptions();
      pmo.options = MQC.MQPMO_SET_IDENTITY_CONTEXT + MQC.MQPMO_FAIL_IF_QUIESCING + MQC.MQPMO_LOGICAL_ORDER + MQC.MQPMO_SYNCPOINT;
      MQFM.logger.info("putMsgGreater4MB: Attempting to send a file '"+so.srcFilename);

      /*
       * MQFM will send the data in chucks of almost 4MB.
       */
      long fileLen = (new File(so.srcFilename)).length();
      long numberOfBlocks = fileLen / Constants.MAX_GROUP_MSG_SIZE;
      int leftOver = (int)(fileLen % Constants.MAX_GROUP_MSG_SIZE);

      byte[] byteData = new byte[Constants.MAX_GROUP_MSG_SIZE];

      try
      {
         MQMessage sendMsg = createMQFMMessage(so);

         sendMsg.messageFlags = MQC.MQMF_MSG_IN_GROUP;
         
         in = new RandomAccessFile(so.srcFilename, "r");
         fileChannel = in.getChannel();
         fileLock = fileChannel.tryLock(0, fileLen, true); /* shared lock for reading */

         if (fileLock == null)
         {
            MQFM.logger.fatal("putMsgGreater4MB cannot acquire lock for file: " + so.srcFilename);
            MQFM.globalErrorText = "putMsgGreater4MB cannot acquire lock for file: " + so.srcFilename;
            throw new MQFMException("putMsgGreater4MB cannot acquire lock for file: " + so.srcFilename);
         }
         
         for (int i=0; i < numberOfBlocks; i++)
         {
            if ( (i == (numberOfBlocks - 1)) && (leftOver == 0) )
               sendMsg.messageFlags = MQC.MQMF_LAST_MSG_IN_GROUP; 
            else
               sendMsg.messageFlags = MQC.MQMF_MSG_IN_GROUP;

            in.read(byteData);
            
            if ( (so.crypto != null) && (so.format == Constants.FORMAT_ZIP) )
            {
               AES aes = new AES(so.crypto.passPhrase, so.crypto.keySize);
               sendMsg.write(aes.encrypt(JZip.compressData(byteData)));
            }
            else if (so.crypto != null)
            {
               AES aes = new AES(so.crypto.passPhrase, so.crypto.keySize);
               sendMsg.write(aes.encrypt(byteData));
            }
            else if (so.format == Constants.FORMAT_ZIP)
               sendMsg.write(JZip.compressData(byteData));
            else
               sendMsg.write(byteData);
            
            fileSize += byteData.length;
            
            if (outDL != null)
               outDL.put(sendMsg, pmo); // put the message to distribution list
            else
               outQ.put(sendMsg, pmo); // put the message on the queue.
            
            so.msgCount++;
            
            MQFM.logger.info("Putting message # "+(i+1)+" in group from file '" + so.srcFilename);
            sendMsg.clearMessage();
            sendMsg.messageId = MQC.MQMI_NONE; // each message should have a unique MsgId.
         }

         if (leftOver != 0) 
         {
            sendMsg.messageFlags = MQC.MQMF_LAST_MSG_IN_GROUP; 
            byteData = new byte[(int)leftOver];
            in.read(byteData);
            
            if ( (so.crypto != null) && (so.format == Constants.FORMAT_ZIP) )
            {
               AES aes = new AES(so.crypto.passPhrase, so.crypto.keySize);
               sendMsg.write(aes.encrypt(JZip.compressData(byteData)));
            }
            else if (so.crypto != null)
            {
               AES aes = new AES(so.crypto.passPhrase, so.crypto.keySize);
               sendMsg.write(aes.encrypt(byteData));
            }
            else if (so.format == Constants.FORMAT_ZIP)
               sendMsg.write(JZip.compressData(byteData));
            else
               sendMsg.write(byteData);
            
            fileSize += byteData.length;
            
            if (outDL != null)
               outDL.put(sendMsg, pmo); // put the message to distribution list
            else
               outQ.put(sendMsg, pmo); // put the message on the queue.
            
            so.msgCount++;
            MQFM.logger.info("Putting message # "+(numberOfBlocks+1)+" (last) in group from file '" + so.srcFilename);
         }
         
         so.fileCount++;
         
         MQFM.logger.info("Put group of messages to queue "+mqo.qObject +" : Filesize is "+fileSize);
         
         _qMgr.commit();
         MQFM.logger.info("Issued commit." );
         
         processingOK = true;
      }
      catch (MQException e)
      {
         MQFM.logger.fatal("MQException: " + e.getLocalizedMessage() + " : QName=" + mqo.qObject);
         MQFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
         MQFM.globalErrorText = e.getLocalizedMessage();
         backout();
         processingOK = false;
      }
      catch (IOException ioex)
      {
         MQFM.logger.fatal("IOException: " + ioex.getLocalizedMessage());
         MQFM.globalErrorText = ioex.getLocalizedMessage();
         backout();
         processingOK = false;
      }
      finally
      {
         try
         {
            if(fileLock != null) 
               fileLock.release();         
         }
         catch (IOException e)
         {
            MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         }
         
         try
         {
            if (fileChannel != null)
               fileChannel.close();
         }
         catch (IOException e)
         {
            MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         }
         
         try
         {
            if (in != null)
               in.close();
         }
         catch (IOException e)
         {
            MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         }
      }
      
      try
      {
         if (processingOK)
            handlePostCommon(so);
      }
      catch (IOException e)
      {
         MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
         MQFM.globalErrorText = e.getLocalizedMessage();
      }
   }

   /**
    * Create an MQMessage for MQFM to use.
    *
    * @param so
    * @throws JDOMException
    * @throws IOException 
    */
   public MQMessage createMQFMMessage(SendObject so) throws JDOMException, IOException
   {
      String remoteFilename;
      String remoteDirectory;
      MQFM.logger.info("Creating MQMessage for MQFM.");

      MQMessage sendMsg = new MQMessage();

      sendMsg.messageId    = MQC.MQMI_NONE;
      sendMsg.correlationId= MQC.MQCI_NONE;
      sendMsg.groupId      = MQC.MQGI_NONE;
      sendMsg.persistence  = (so.persistence? MQC.MQPER_PERSISTENT : MQC.MQPER_NOT_PERSISTENT);
      sendMsg.priority     = so.priority;
      sendMsg.messageType  = MQC.MQMT_DATAGRAM;
      sendMsg.format       = MQC.MQFMT_NONE;
      sendMsg.feedback     = Constants.MQFM_MSG;

      if (mqo.userID != null)
         sendMsg.userId = mqo.userID;
      else
         sendMsg.userId = System.getProperty("user.name");

      if ( (so.crypto != null) && (so.format == Constants.FORMAT_ZIP) )
         sendMsg.feedback = Constants.MQFM_ZIP_ENCRYPT_MSG;
      else if (so.crypto != null)
         sendMsg.feedback = Constants.MQFM_ENCRYPT_MSG;
      else if (so.format == Constants.FORMAT_ZIP)
         sendMsg.feedback = Constants.MQFM_ZIP_MSG;
      else if (so.format == Constants.FORMAT_STRING)
         sendMsg.format   = MQC.MQFMT_STRING;

      remoteFilename = Tokens.replace(so.actionName, so.remoteFilename, so.srcFilename);

      remoteDirectory = Tokens.replace(so.actionName,
                                       so.remoteDirectory,
                                       IO.getPath(so.srcFilename) );

      /**
       * Set MQMD Context fields in the data fits or 
       * create XML and put the XML in the payload.
       */
      if ( (so.remoteFilename.length() > MQC.MQ_APPL_IDENTITY_DATA_LENGTH)   ||
           (so.remoteDirectory.length() > MQC.MQ_ACCOUNTING_TOKEN_LENGTH)    ||
           ( (so.remoteJobXmlFile != null) &&
             (so.remoteJobXmlFile.length() > MQC.MQ_CORREL_ID_LENGTH) ) )
      {
         sendMsg.feedback = sendMsg.feedback | Constants.MQFM_XML_EMBEDDED_MSG;

         StringBuffer xmlData = new StringBuffer();
         xmlData.append(Constants.XML_PROLOG);
         xmlData.append("<" + Constants.MQFM_MSG_ELEMENT + ">");
         xmlData.append("<" + Constants.REMOTE_ELEMENT + ">");
         if (so.remoteJobXmlFile != null)
         {
            xmlData.append("<" + Constants.EXECUTE_ELEMENT + " " + Constants.XML_FILE_ATTRIBUTE + "=\""  + so.remoteJobXmlFile + "\" />" );
            sendMsg.feedback = sendMsg.feedback | Constants.MQFM_START_JOB;
         }
         xmlData.append("<" + Constants.DIRECTORY_ELEMENT + ">" + remoteDirectory +"</" + Constants.DIRECTORY_ELEMENT + ">");
         xmlData.append("<" + Constants.FILE_ELEMENT + ">" + remoteFilename + "</" + Constants.FILE_ELEMENT + ">");
         xmlData.append("</" + Constants.REMOTE_ELEMENT + ">");
         xmlData.append("</" + Constants.MQFM_MSG_ELEMENT + ">");
         byte[] xmlByteData = xmlData.toString().getBytes();

         /* save the XML data length */
         int len = xmlByteData.length;
         sendMsg.correlationId[0] = (byte) ((len >>> 24) & 0xffL);
         sendMsg.correlationId[1] = (byte) ((len >>> 16) & 0xffL);
         sendMsg.correlationId[2] = (byte) ((len >>> 8)  & 0xffL);
         sendMsg.correlationId[3] = (byte) ((len) & 0xffL);

         sendMsg.write(xmlByteData);
      }
      else
      {
         sendMsg.applicationIdData = so.remoteFilename;
         so.remoteDirectory += "                                ";  // pad it with blanks.
         for (int i = 0; (i < so.remoteDirectory.length()) && (i < MQC.MQ_ACCOUNTING_TOKEN_LENGTH); i++)
            sendMsg.accountingToken[i] = (byte) so.remoteDirectory.charAt(i);

         if (so.remoteJobXmlFile != null)
         {
            String remoteJobFile;
            sendMsg.feedback = sendMsg.feedback | Constants.MQFM_START_JOB;
            if (so.remoteJobXmlFile.endsWith(Constants.XML_SUFFIX))
            { /* don't send the ".xml" part of the filename */
               int xx = so.remoteJobXmlFile.lastIndexOf(Constants.XML_SUFFIX);
               remoteJobFile = so.remoteJobXmlFile.substring(0,xx);
            }
            else
               remoteJobFile = so.remoteJobXmlFile;

            for (int i = 0; (i < remoteJobFile.length()) && (i < MQC.MQ_CORREL_ID_LENGTH); i++)
               sendMsg.correlationId[i] = (byte) remoteJobFile.charAt(i);
         }
      }
      
      return sendMsg;
   }

   /**
    * Handle the post common code.
    * @param so
    * @throws JDOMException
    * @throws IOException
    * @throws MQFMException
    */
   private void handlePostCommon(SendObject so) throws JDOMException, IOException, MQFMException
   {
      if (so.jobXmlFile != null)
      {
         RunJob rj = new RunJob();
         rj.launch(so.jobXmlFile);
      }

      if (so.archiveFile)
      {
         String archiveDirectory = Tokens.replace(so.archiveDirectory, MQFM.userProperties);
         String archiveFileName = null;
         if (so.archiveFileName != null)
            archiveFileName = Tokens.replace(so.actionName, so.archiveFileName, IO.getFileName(so.srcFilename) );
         else
            archiveFileName = IO.getFileName(so.srcFilename);

         handleDirectoryCheck(so.archiveCreateDir, archiveDirectory);
         
         IO.archiveFile(so.srcFilename, archiveDirectory, archiveFileName);
      }
      else if (so.deleteFile)
         IO.deleteFile(so.srcFilename);
   }
   
   /**
    * Create the archive directory.
    * @param archiveCreateDir
    * @param archiveDirectory
    * @throws IOException
    */
   private void handleDirectoryCheck(boolean archiveCreateDir, String archiveDirectory) throws IOException
   {
      /* if it does not exist then create it */
      if (!(new File(archiveDirectory).exists()))
      {
         if (archiveCreateDir)
         {
            if ((new File(archiveDirectory)).mkdirs()) 
            {
               MQFM.logger.info("Successfully created archive directory: " + archiveDirectory);
            }
            else
            {
               String errorMsg = "Cannot create archive directory: " + archiveDirectory;
               MQFM.logger.fatal(errorMsg);
               MQFM.globalErrorText = errorMsg;
               new IOException("createArchiveDirectory: "+errorMsg);
            }
         }
      
      }
   }
}
