package biz.capitalware.mqfm.schedule;

import java.util.Calendar;
import java.util.StringTokenizer;

import biz.capitalware.mqfm.MQFM;

/**
 * This class holds the values for the Schedule action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class Schedule
{
   public String  xmlFile = null;
   public int[]   minutes = new int[]{-1};
   public int[]   hours = new int[]{-1};
   public int[]   daysOfMonth = new int[]{-1};
   public int[]   months = new int[]{-1};
   public int[]   daysOfWeek = new int[]{-1};
   public int     year = -1;

   /**
    * The constructor
    */
   public Schedule()
   {
      super();
   }

   /**
    * Verify the values in the object.
    */
   public void checkObject()
   {
      int i;
      
      if (xmlFile == null)
         throw new IllegalArgumentException("Error: You must specify an MQFM Workflow XML filename for the Sechdule action.");
      
      for (i=0; i < minutes.length; i++)
      {
         if ( (minutes[i] < -1) || (minutes[i] > 59) )
            throw new IllegalArgumentException("Error: The minute value is invalid. Use '*' or a number between 0 and 59.");
      }

      for (i=0; i < hours.length; i++)
      {
         if ( (hours[i] < -1) || (hours[i] > 23) )
            throw new IllegalArgumentException("Error: The hour value is invalid. Use '*' or a number between 0 and 23.");
      }

      for (i=0; i < daysOfMonth.length; i++)
      {
         if ( (daysOfMonth[i] < -1) || (daysOfMonth[i] == 0) || (daysOfMonth[i] > 31) )
            throw new IllegalArgumentException("Error: The DayOfMonth value is invalid. Use '*' or a number between 1 and 31.");
      }

      for (i=0; i < months.length; i++)
      {
         if ( (months[i] < -1) || (months[i] == 0) || (months[i] > 12) )
            throw new IllegalArgumentException("Error: The month value is invalid. Use '*' or a number between 1 and 12.");
      }

      for (i=0; i < daysOfWeek.length; i++)
      {
         if ( (daysOfWeek[i] < -1) || (daysOfWeek[i] > 6) )
            throw new IllegalArgumentException("Error: The dayOfWeek value is invalid. Use '*' or use MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY or SUNDAY.");
      }

      Calendar c = Calendar.getInstance();
      if ( (year < -1) || ( (year >= 0) && (year < c.get(Calendar.YEAR)) ) )
         throw new IllegalArgumentException("Error: The year value is invalid. Use '*' or a number equal to or higher than the current year.");
   }

   /**
    * Output the info
    * @param logger
    */
   public void printValues()
   {
      MQFM.logger.info("     MQFM Workflow XML    = " + xmlFile);
      MQFM.logger.info("     Minute               = " + formatIntsToString(minutes));
      MQFM.logger.info("     Hour                 = " + formatIntsToString(hours));
      MQFM.logger.info("     DayOfMonth           = " + formatIntsToString(daysOfMonth));
      MQFM.logger.info("     Month                = " + formatIntsToMonth(months));
      MQFM.logger.info("     DayOfWeek            = " + formatIntsToDayOfWeek(daysOfWeek));
      MQFM.logger.info("     Year                 = " + (year == -1 ? "*" : year));
   }

   /**
    * Convert an array of "int"s to an array of Strings
    * @param value
    * @return 
    */
   private String formatIntsToString(int[] value)
   {
      int i;
      StringBuffer sb = new StringBuffer(); 
      
      for (i=0; i < value.length; i++)
      {
         if (value[i] == -1)
            sb.append("*,");
         else
            sb.append(value[i]+",");
      }
      
      sb.deleteCharAt(sb.length()-1); // remove the last ","
      
      return sb.toString();
   }

   /**
    * Convert an array of "int"s to an array of Strings
    * @param value
    * @return 
    */
   private String formatIntsToDayOfWeek(int[] value)
   {
      int i;
      StringBuffer sb = new StringBuffer(); 
      
      for (i=0; i < value.length; i++)
      {
         if (value[i] == -1)
            sb.append("*,");
         else
         {
            if (value[i] == 2)
               sb.append("MONDAY,");
            else if (value[i] == 3)
               sb.append("TUESDAY,");
            else if (value[i] == 4)
               sb.append("WEDNESDAY,");
            else if (value[i] == 5)
               sb.append("THURSDAY,");
            else if (value[i] == 6)
               sb.append("FRIDAY,");
            else if (value[i] == 7)
               sb.append("SATURDAY,");
            else if (value[i] == 1)
               sb.append("SUNDAY,");
         }
      }
      
      sb.deleteCharAt(sb.length()-1); // remove the last ","
      
      return sb.toString();
   }

   /**
    * Convert an array of "int"s to an array of Strings
    * @param value
    * @return 
    */
   private String formatIntsToMonth(int[] value)
   {
      int i;
      StringBuffer sb = new StringBuffer(); 
      
      for (i=0; i < value.length; i++)
      {
         if (value[i] == -1)
            sb.append("*,");
         else
         {
            if (value[i] == 0)
               sb.append("JANUARY,");
            else if (value[i] == 1)
               sb.append("FEBRUARY,");
            else if (value[i] == 2)
               sb.append("MARCH,");
            else if (value[i] == 3)
               sb.append("APRIL,");
            else if (value[i] == 4)
               sb.append("MAY,");
            else if (value[i] == 5)
               sb.append("JUNE,");
            else if (value[i] == 6)
               sb.append("JULY,");
            else if (value[i] == 7)
               sb.append("AUGUST,");
            else if (value[i] == 8)
               sb.append("SEPTEMBER,");
            else if (value[i] == 9)
               sb.append("OCTOBER,");
            else if (value[i] == 10)
               sb.append("NOVEMBER,");
            else if (value[i] == 11)
               sb.append("DECEMBER,");
         }
      }
      
      sb.deleteCharAt(sb.length()-1); // remove the last ","
      
      return sb.toString();
   }

   /**
    * Convert a string to an int.
    * @param value
    * @return
    */
   public int formatStringToInt(String value)
   {
      int n = -1;
      if ("*".equals(value))
         n = -1;
      else if ( ("".equals(value)) || (value == null) )
         n = -1;
      else
         n = Integer.parseInt(value);
      
      return n;
   }

   /**
    * Convert a string to an array of "int"s
    * @param value
    * @return
    */
   public int[] formatStringToInts(String value)
   {
      int[] nn;
      int i=0;
      if ("*".equals(value))
         nn = new int[]{-1};
      else if ( ("".equals(value)) || (value == null) )
         nn = new int[]{-1};
      else
      {
         StringTokenizer st = new StringTokenizer(value,",");
         nn = new int[st.countTokens()];
         
         if (nn.length > 0)
         {
            while (st.hasMoreTokens())
               nn[i++] = Integer.parseInt(st.nextToken());
         }
         else
            nn = new int[]{-1};
      }
      
      return nn;
   }

   /**
    * Convert a day of week to an array of "int"s
    * @param value
    * @return
    */
   public int[] formatDayOfWeekToInts(String value)
   {
      int[] nn;
      int i=0;
      if ("*".equals(value))
         nn = new int[]{-1};
      else if ( ("".equals(value)) || (value == null) )
         nn = new int[]{-1};
      else
      {
         String day;
         StringTokenizer st = new StringTokenizer(value,",");
         nn = new int[st.countTokens()];
         
         if (nn.length > 0)
         {
            while (st.hasMoreTokens())
            {
               day = st.nextToken();
               
               if ( ("1".equalsIgnoreCase(day)) || ("MO".equalsIgnoreCase(day)) || ("MONDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 2;
               else if ( ("2".equalsIgnoreCase(day)) || ("TU".equalsIgnoreCase(day)) || ("TUESDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 3;
               else if ( ("3".equalsIgnoreCase(day)) || ("WE".equalsIgnoreCase(day)) || ("WEDNESDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 4;
               else if ( ("4".equalsIgnoreCase(day)) || ("TH".equalsIgnoreCase(day)) || ("THURSDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 5;
               else if ( ("5".equalsIgnoreCase(day)) || ("FR".equalsIgnoreCase(day)) || ("FRIDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 6;
               else if ( ("6".equalsIgnoreCase(day)) || ("SA".equalsIgnoreCase(day)) || ("SATURDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 7;
               else if ( ("0".equalsIgnoreCase(day)) || ("7".equalsIgnoreCase(day)) || ("SU".equalsIgnoreCase(day)) || ("SUNDAY".equalsIgnoreCase(day)) )
                  nn[i++] = 1;
               else
                  nn[i++] = 99; // errror!!
            }
         }
         else
            nn = new int[]{-1};
      }
      
      return nn;
   }

   /**
    * Convert a day of week to an array of "int"s
    * @param value
    * @return
    */
   public int[] formatMonthToInts(String value)
   {
      int[] nn;
      int i=0;
      if ("*".equals(value))
         nn = new int[]{-1};
      else if ( ("".equals(value)) || (value == null) )
         nn = new int[]{-1};
      else
      {
         String month;
         StringTokenizer st = new StringTokenizer(value,",");
         nn = new int[st.countTokens()];
         
         if (nn.length > 0)
         {
            while (st.hasMoreTokens())
            {
               month = st.nextToken();
               
               if ( ("1".equalsIgnoreCase(month)) || ("JAN".equalsIgnoreCase(month)) || ("JANUARY".equalsIgnoreCase(month)) )
                  nn[i++] = 0;
               else if ( ("2".equalsIgnoreCase(month)) || ("FEB".equalsIgnoreCase(month)) || ("FEBRUARY".equalsIgnoreCase(month)) )
                  nn[i++] = 1;
               else if ( ("3".equalsIgnoreCase(month)) || ("MAR".equalsIgnoreCase(month)) || ("MARCH".equalsIgnoreCase(month)) )
                  nn[i++] = 2;
               else if ( ("4".equalsIgnoreCase(month)) || ("APR".equalsIgnoreCase(month)) || ("APRIL".equalsIgnoreCase(month)) )
                  nn[i++] = 3;
               else if ( ("5".equalsIgnoreCase(month)) || ("MAY".equalsIgnoreCase(month)) )
                  nn[i++] = 4;
               else if ( ("6".equalsIgnoreCase(month)) || ("JUN".equalsIgnoreCase(month)) || ("JUNE".equalsIgnoreCase(month)) )
                  nn[i++] = 5;
               else if ( ("7".equalsIgnoreCase(month)) || ("JUL".equalsIgnoreCase(month)) || ("JULY".equalsIgnoreCase(month)) )
                  nn[i++] = 6;
               else if ( ("8".equalsIgnoreCase(month)) || ("AUG".equalsIgnoreCase(month)) || ("AUGUST".equalsIgnoreCase(month)) )
                  nn[i++] = 7;
               else if ( ("9".equalsIgnoreCase(month)) || ("SEP".equalsIgnoreCase(month)) || ("SEPTEMBER".equalsIgnoreCase(month)) )
                  nn[i++] = 8;
               else if ( ("10".equalsIgnoreCase(month)) || ("OCT".equalsIgnoreCase(month)) || ("OCTOBER".equalsIgnoreCase(month)) )
                  nn[i++] = 9;
               else if ( ("11".equalsIgnoreCase(month)) || ("NOV".equalsIgnoreCase(month)) || ("NOVEMBER".equalsIgnoreCase(month)) )
                  nn[i++] = 10;
               else if ( ("12".equalsIgnoreCase(month)) || ("DEC".equalsIgnoreCase(month)) || ("DECEMBER".equalsIgnoreCase(month)) )
                  nn[i++] = 11;
               else
                  nn[i++] = 99; // errror!!
            }
         }
         else
            nn = new int[]{-1};
      }
      
      return nn;
   }
}
