package biz.capitalware.mqfm.schedule;

import java.util.Calendar;
import java.util.Date;
import java.util.Arrays;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.objects.PastDateException;

/**
 * This class represents the attributes of an event.
 * 
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ScheduleEntry implements Comparable, java.io.Serializable
{
   private int[]        minutes = { -1 };
   private static int   minMinute = 0;
   private static int   maxMinute = 59;

   private int[]        hours = { -1 };
   private static int   minHour = 0;
   private static int   maxHour = 23;

   private int[]        daysOfMonth = { -1 };
   private static int   minDayOfMonth = 1;
   // maxDayOfMonth varies by month

   private int[]        months = { -1 };
   private static int   minMonth = 0;
   private static int   maxMonth = 11;

   private int[]        daysOfWeek = { -1 };
   private static int   minDayOfWeek = 1;
   private static int   maxDayOfWeek = 7;

   private int          year = -1; // no support for a list of years -- must be * or specified

   private String       file;;
   private String       name;
   private static int   UNIQUE = 0; // used to generate names if they are null

   private boolean      notificationInNewThread = false; // default: false

   private boolean      isRelative;
   public boolean       isRepeating;
   public long          eventTime;
   private long         lastUpdateTime;
   private transient ScheduleEventListener listener;

   /**
    * Creates a new EventEntry. Fixed date format: this event will happen once,
    * at the timestamp given.
    * 
    * @param date
    *           the event date to be added.
    * @param file
    *           MQFM Workflow XML file associated with this entry           
    * @param listener
    *           the event listener.
    * @exception PastDateException
    *               if the event date is in the past (or less than 1 second away
    *               from the current date).
    */
   public ScheduleEntry(String _name, 
                     Date _date,
                     String   _file,
                     ScheduleEventListener _listener)
         throws PastDateException
   {
      setName(_name);
      file = _file;
      listener = _listener;
      Calendar event = Calendar.getInstance();
      event.setTime(_date);
      minutes = new int[] { event.get(Calendar.MINUTE) };
      hours = new int[] { event.get(Calendar.HOUR_OF_DAY) };
      daysOfMonth = new int[] { event.get(Calendar.DAY_OF_MONTH) };
      months = new int[] { event.get(Calendar.MONTH) };
      year = event.get(Calendar.YEAR);

      isRepeating = false;
      isRelative = false;
      eventTime = _date.getTime();
      checkScheduleTime();
   }

   /**
    * Creates a new EventEntry. Delay format: this event will happen once or
    * repeatedly, at increments of the number of minutes given.
    * 
    * @param _name
    *           keeps the event unique from other events with the same schedule,
    *           and used for debugging.
    * @param delayMinutes
    *           the event delay in minutes (relative to now).
    * @param isRepetitive
    *           <code>true</code> if the event must be reactivated,
    *           <code>false</code> otherwise.
    * @param file
    *           MQFM Workflow XML file associated with this entry           
    * @param listener
    *           the event listener.
    * @exception PastDateException
    *               if the event date is in the past (or less than 1 second
    *               closed to the current date).
    */
   public ScheduleEntry(String _name, 
                     int _delayMinutes, 
                     boolean _isRepeating,
                     String   _file,
                     ScheduleEventListener _listener) 
      throws PastDateException
   {
      if (_delayMinutes < 1)
      {
         throw new PastDateException();
      }

      setName(_name);
      file = _file;
      minutes = new int[] { _delayMinutes };
      listener = _listener;
      isRepeating = _isRepeating;

      isRelative = true;
      updateScheduleTime();
   }

   /**
    * <p>
    * Creates a new EventEntry. Basic cron format - use each field to restrict
    * events to a specific minute, hour, etc. OR pass in -1 to allow all values
    * of that field.
    * </p>
    * 
    * <p>
    * Params of (30, 13, -1, -1, 2, -1, listener) schedule an event for 1:30pm
    * every Monday.
    * </p>
    * 
    * <p>
    * NOTE: if both dayOfMonth and dayOfWeek are restricted, each event will be
    * scheduled for the sooner match.
    * </p>
    * 
    * @param minute
    *           minute of the event. Allowed values 0-59.
    * @param hour
    *           hour of the event. Allowed values 0-23.
    * @param dayOfMonth
    *           day of month of the event (-1 if every day). Allowed values
    *           1-31.
    * @param month
    *           month of the event (-1 if every month). Allowed values 0-11 (0 =
    *           January, 1 = February, ...). <code>java.util.Calendar</code>
    *           constants can be used.
    * @param dayOfWeek
    *           day of week of the event (-1 if every day). This attribute is
    *           exclusive with <code>dayOfMonth</code>. Allowed values 1-7 (1 =
    *           Sunday, 2 = Monday, ...). <code>java.util.Calendar</code>
    *           constants can be used.
    * @param year
    *           year of the event. When this field is not set (i.e. -1) the
    *           event is repetitive (i.e. it is rescheduled when reached).
    * @param file
    *           MQFM Workflow XML file associated with this entry           
    * @param listener
    *           the event listener.
    * @return the EventEntry.
    * @exception PastDateException
    *               if the event date is in the past (or less than 1 second away
    *               from the current date).
    */
   public ScheduleEntry(String _name, 
                     int _minute, 
                     int _hour, 
                     int _dayOfMonth,
                     int _month, 
                     int _dayOfWeek, 
                     int _year,
                     String   _file,
                     ScheduleEventListener _listener)
      throws PastDateException
   {
      this(_name, 
           new int[] { _minute }, 
           new int[] { _hour },
           new int[] { _dayOfMonth }, 
           new int[] { _month },
           new int[] { _dayOfWeek }, 
           _year,
           _file,
           _listener);
   }

   /**
    * <p>
    * Creates a new EventEntry. Extended cron format - supports lists of values
    * for each field, or {-1} to allow all values for that field.
    * </p>
    * 
    * <p>
    * Params of (30, 13, -1, -1, 2, -1, listener) schedule an event for 1:30pm
    * every Monday.
    * </p>
    * 
    * <p>
    * NOTE: if both dayOfMonth and dayOfWeek are restricted, each event will be
    * scheduled for the sooner match.
    * </p>
    * 
    * @param minutes
    *           valid minutes of the event. Allowed values 0-59, or {-1} for
    *           all.
    * @param hours
    *           valid hours of the event. Allowed values 0-23, or {-1} for all.
    * @param daysOfMonth
    *           valid days of month of the event. Allowed values 1-31, or {-1}
    *           for all.
    * @param months
    *           valid months of the event. Allowed values 0-11 (0 = January, 1 =
    *           February, ...), or {-1} for all. <code>java.util.Calendar</code>
    *           constants can be used.
    * @param daysOfWeek
    *           valid days of week of the event. This attribute is exclusive
    *           with <code>dayOfMonth</code>. Allowed values 1-7 (1 = Sunday, 2
    *           = Monday, ...), or {-1} for all. <code>java.util.Calendar</code>
    *           constants can be used.
    * @param year
    *           year of the event. When this field is not set (i.e. -1) the
    *           event is repetitive (i.e. it is rescheduled when reached).
    * @param file
    *           MQFM Workflow XML file associated with this entry           
    * @param listener
    *           the event listener.
    * @return the EventEntry.
    * @exception PastDateException
    *               if the event date is in the past (or less than 1 second away
    *               from the current date).
    */
   public ScheduleEntry(String   _name, 
                     int[]    _minutes, 
                     int[]    _hours,
                     int[]    _daysOfMonth, 
                     int[]    _months, 
                     int[]    _daysOfWeek, 
                     int      _year,
                     String   _file,
                     ScheduleEventListener _listener) 
      throws PastDateException
   {
      setName(_name);
      file = _file;
      minutes = _minutes;
      hours = _hours;
      daysOfMonth = _daysOfMonth;
      months = _months;
      daysOfWeek = _daysOfWeek;
      year = _year;
      listener = _listener;
      isRepeating = (_year == -1);
      isRelative = false;

      updateScheduleTime();
      checkScheduleTime();
   }

   /**
    * Just make sure it's not null -- and if it is, make it unique.
    * 
    * @param _name
    */
   private void setName(String _name)
   {
      name = _name;
      if (name == null)
         name = "event" + (UNIQUE++);
   }

   public String getName()
   {
      return name;
   }

   /**
    * Set the associated file name
    *
    * @param _file
    */
   public void setFile(String _file)
   {
      file = _file;
   }

   /**
    * Retrieve the file name
    *
    * @return file
    */
   public String getFile()
   {
      return file;
   }

   /**
    * By default, the ScheduleEventListeners for all events will be notified in
    * the same thread (so a long-running handleEvent() implementation will cause
    * other events to wait until it completes). Call this method to notify the
    * listener to this event in a new Thread, so other events won't be delayed.
    */
   public void setNotificationInNewThead()
   {
      notificationInNewThread = true;
   }

   public boolean isNotificationInNewThread()
   {
      return notificationInNewThread;
   }

   /**
    * Checks that event is not in the past, or less than 1 second away.
    * 
    * @exception PastDateException
    *               if the event date is in the past (or less than 1 second in
    *               the future).
    */
   void checkScheduleTime() throws PastDateException
   {
      long delay = eventTime - System.currentTimeMillis();

      if (delay <= 1000)
      {
         throw new PastDateException();
      }
   }

   /**
    * Notifies the listener.
    */
   public void scheduleNotification()
   {
      listener.handleEvent(this);
   }

   /**
    * Updates this event entry to the next valid event time, AFTER the current
    * time.
    */
   public void updateScheduleTime()
   {
      Calendar now = Calendar.getInstance();

      if (isRelative)
      {
         // relative only uses minutes field, with only a single value (NOT -1)
         eventTime = now.getTime().getTime() + (minutes[0] * 60000);
         return;
      }

      Calendar event = (Calendar) now.clone();
      event.set(Calendar.SECOND, 0);

      MQFM.logger.debug("now: " + now.getTime());

      //
      // the updates work in a cascade -- if next minute value is in the
      // following hour, hour is incremented. If next valid hour value is
      // in the following day, day is incremented, and so on.
      //

      // increase event minutes
      int current = event.get(Calendar.MINUTE);
      int offset = 0;
      // force increment at least to next minute
      offset = getOffsetToNext(current, minMinute, maxMinute, minutes);
      event.add(Calendar.MINUTE, offset);
      MQFM.logger.debug("after min: " + event.getTime());

      // update event hours if necessary
      current = event.get(Calendar.HOUR_OF_DAY); // (as updated by minute shift)
      offset = getOffsetToNextOrEqual(current, minHour, maxHour, hours);
      event.add(Calendar.HOUR_OF_DAY, offset);
      MQFM.logger.debug("after hour (current:" + current + "): " + event.getTime());

      //
      // If days of month AND days of week are restricted, we take whichever
      // match
      // comes sooner.
      // If only one is restricted, take the first match for that one.
      // If neither is restricted, don't do anything.
      //
      if (daysOfMonth[0] != -1 && daysOfWeek[0] != -1)
      {
         // BOTH are restricted - take earlier match
         Calendar dayOfWeekEvent = (Calendar) event.clone();
         updateDayOfWeekAndMonth(dayOfWeekEvent);

         Calendar dayOfMonthEvent = (Calendar) event.clone();
         updateDayOfMonthAndMonth(dayOfMonthEvent);

         // take the earlier one
         if (dayOfMonthEvent.getTime().getTime() < dayOfWeekEvent.getTime()
               .getTime())
         {
            event = dayOfMonthEvent;
            MQFM.logger.debug("after dayOfMonth CLOSER: " + event.getTime());
         }
         else
         {
            event = dayOfWeekEvent;
            MQFM.logger.debug("after dayOfWeek CLOSER: " + event.getTime());
         }
      }
      else if (daysOfWeek[0] != -1) // only dayOfWeek is restricted
      {
         // update dayInWeek and month if necessary
         updateDayOfWeekAndMonth(event);
         MQFM.logger.debug("after dayOfWeek: " + event.getTime());
      }
      else if (daysOfMonth[0] != -1) // only dayOfMonth is restricted
      {
         // update dayInMonth and month if necessary
         updateDayOfMonthAndMonth(event);
         MQFM.logger.debug("after dayOfMonth: " + event.getTime());
      }
      // else if neither is restricted (both[0] == -1), we don't need to do
      // anything.

      MQFM.logger.debug("event: " + event.getTime());

      eventTime = event.getTime().getTime();
      lastUpdateTime = System.currentTimeMillis();
   }

   /**
    * daysInMonth can't use simple offsets like the other fields, because the
    * number of days varies per month (think of an event that executes on every
    * 31st). Instead we advance month and dayInMonth together until we're on a
    * matching value pair.
    */
   void updateDayOfMonthAndMonth(Calendar event)
   {
      int currentMonth = event.get(Calendar.MONTH);
      int currentDayOfMonth = event.get(Calendar.DAY_OF_MONTH);
      int offset = 0;

      // loop until we have a valid day AND month (if current is invalid)
      while (!isIn(currentMonth, months)
            || !isIn(currentDayOfMonth, daysOfMonth))
      {
         // if current month is invalid, advance to 1st day of next valid month
         if (!isIn(currentMonth, months))
         {
            offset = getOffsetToNextOrEqual(currentMonth, minMonth, maxMonth,
                  months);
            event.add(Calendar.MONTH, offset);
            event.set(Calendar.DAY_OF_MONTH, 1);
            currentDayOfMonth = 1;
         }

         // advance to the next valid day of month, if necessary
         if (!isIn(currentDayOfMonth, daysOfMonth))
         {
            int maxDayOfMonth = event.getActualMaximum(Calendar.DAY_OF_MONTH);
            offset = getOffsetToNextOrEqual(currentDayOfMonth, minDayOfMonth,
                  maxDayOfMonth, daysOfMonth);
            event.add(Calendar.DAY_OF_MONTH, offset);
         }

         currentMonth = event.get(Calendar.MONTH);
         currentDayOfMonth = event.get(Calendar.DAY_OF_MONTH);
      }
   }

   void updateDayOfWeekAndMonth(Calendar event)
   {
      int currentMonth = event.get(Calendar.MONTH);
      int currentDayOfWeek = event.get(Calendar.DAY_OF_WEEK);
      int offset = 0;

      // loop until we have a valid day AND month (if current is invalid)
      while (!isIn(currentMonth, months) || !isIn(currentDayOfWeek, daysOfWeek))
      {
         // if current month is invalid, advance to 1st day of next valid month
         if (!isIn(currentMonth, months))
         {
            offset = getOffsetToNextOrEqual(currentMonth, minMonth, maxMonth,
                  months);
            event.add(Calendar.MONTH, offset);
            event.set(Calendar.DAY_OF_MONTH, 1);
            currentDayOfWeek = event.get(Calendar.DAY_OF_WEEK);
         }

         // advance to the next valid day of week, if necessary
         if (!isIn(currentDayOfWeek, daysOfWeek))
         {
            offset = getOffsetToNextOrEqual(currentDayOfWeek, minDayOfWeek,
                  maxDayOfWeek, daysOfWeek);
            event.add(Calendar.DAY_OF_YEAR, offset);
         }

         currentDayOfWeek = event.get(Calendar.DAY_OF_WEEK);
         currentMonth = event.get(Calendar.MONTH);
      }
   }

   // ----------------------------------------------------------------------
   // General utility methods
   // ----------------------------------------------------------------------

   /**
    * if values = {-1} offset is 1 (because next value definitely matches) if
    * current < last(values) offset is diff to next valid value if current >=
    * last(values) offset is diff to values[0], wrapping from max to min
    */
   static int getOffsetToNext(int current, int min, int max, int[] values)
   {
      int offset = 0;

      // find the distance to the closest valid value > current (wrapping if
      // neccessary)

      // {-1} means * -- offset is 1 because current++ is valid value
      if (values[0] == -1)
      {
         offset = 1;
      }
      else
      {
         // need to wrap
         if (current >= last(values))
         {
            int next = values[0];
            offset = (max - current + 1) + (next - min);
         }
         else
         // current < max(values) -- find next valid value after current
         {
            findvalue: for (int i = 0; i < values.length; i++)
            {
               if (current < values[i])
               {
                  offset = values[i] - current;
                  break findvalue;
               }
            }
         } // end current < max(values)
      }

      return offset;
   }

   /**
    * if values = {-1} or current is valid offset is 0. if current <
    * last(values) offset is diff to next valid value if current >= last(values)
    * offset is diff to values[0], wrapping from max to min
    */
   static int getOffsetToNextOrEqual(int current, int min, int max, int[] values)
   {
      int offset = 0;
      int[] safeValues = null;

      // find the distance to the closest valid value >= current (wrapping if
      // necessary)

      // {-1} means * -- offset is 0 if current is valid value
      if (values[0] == -1 || isIn(current, values))
      {
         offset = 0;
      }
      else
      {
         safeValues = discardValuesOverMax(values, max);

         // need to wrap
         if (current > last(safeValues))
         {
            int next = safeValues[0];
            offset = (max - current + 1) + (next - min);
         }
         else
         // current <= max(values) -- find next valid value
         {
            findvalue: for (int i = 0; i < values.length; i++)
            {
               if (current < safeValues[i])
               {
                  offset = safeValues[i] - current;
                  break findvalue;
               }
            }
         } // end current <= max(values)
      }

      return offset;
   }

   /**
    * handles -1 in values as * and returns true otherwise returns true iff
    * given value is in the array
    */
   static boolean isIn(int find, int[] values)
   {
      if (values[0] == -1)
      {
         return true;
      }
      else
      {
         for (int i = 0; i < values.length; i++)
         {
            if (find == values[i])
               return true;
         }
         return false;
      }
   }

   /**
    * @return the last int in the array
    */
   static int last(int[] intArray)
   {
      return intArray[intArray.length - 1];
   }

   /**
    * Assumes inputted values are not null, have at least one value, and are in
    * ascending order.
    * 
    * @return copy of values without any trailing values that exceed the max
    */
   static int[] discardValuesOverMax(int[] values, int max)
   {
      int[] safeValues = null;
      for (int i = 0; i < values.length; i++)
      {
         if (values[i] > max)
         {
            safeValues = new int[i];
            System.arraycopy(values, 0, safeValues, 0, i);
            return safeValues;
         }
      }
      return values;
   }

   private static String arrToString(int[] intArray)
   {
      if (intArray == null)
         return "null";
      if (intArray.length == 0)
         return "{}";

      String s = "{";
      for (int i = 0; i < intArray.length - 1; i++)
      {
         s += intArray[i] + ", ";
      }
      s += intArray[intArray.length - 1] + "}";

      return s;
   }

   // ----------------------------------------------------------------------
   // Comparable interface
   // ----------------------------------------------------------------------

   /**
    * Compares this EventEntry with the specified EventEntry for order. One
    * twist -- if the eventTime matches, this event will STILL place itself
    * before the other based on the lastUpdateTime. If the other event has been
    * rung more recently, this one should get priority.
    * 
    * @param obj
    *           the EventEntry with which to compare.
    * @return a negative integer, zero, or a positive integer as this EventEntry
    *         is less than, equal to, or greater than the given EventEntry.
    * @exception ClassCastException
    *               if the specified Object's type prevents it from being
    *               compared to this EventEntry.
    */
   public int compareTo(Object obj)
   {
      int flag = 0;

      ScheduleEntry other = (ScheduleEntry) obj;
      if (eventTime < other.eventTime)
         flag = -1;
      else if (eventTime > other.eventTime)
         flag = 1;
      else
      // eventTime == other.eventTime
      {
         if (lastUpdateTime < other.lastUpdateTime)
            flag = -1;
         else if (lastUpdateTime > other.lastUpdateTime)
            flag = 1;
         else
         {
            int k = name.compareTo(other.getName());

            if (k < 0)
               flag = -1;
            else if (k > 0)
               flag = 1;
            else
               flag = 0;
         }
      }

      // System.err.println("compareTo flag="+flag);
      return flag;
   }

   /**
    * Indicates whether some other EventEntry is "equal to" this one. This is
    * where the name is important, since two events can have the exact same
    * schedule.
    * 
    * @param obj
    *           the EventEntry with which to compare.
    * @return <code>true if this EventEntry has the same name,
    * <code>eventTime</code> AND the same schedule as the obj argument;
    *         <code>false</code> otherwise.
    */
   public boolean equals(Object obj)
   {
      ScheduleEntry entry = null;

      if (obj == null || !(obj instanceof ScheduleEntry))
         return false;

      entry = (ScheduleEntry) obj;
      return (name.equals(entry.name) && eventTime == entry.eventTime
            && isRelative == entry.isRelative
            && isRepeating == entry.isRepeating
            && Arrays.equals(minutes, entry.minutes)
            && Arrays.equals(hours, entry.hours)
            && Arrays.equals(daysOfMonth, entry.daysOfMonth)
            && Arrays.equals(months, entry.months) && Arrays.equals(daysOfWeek,
            entry.daysOfWeek));
   }

   /**
    * @return a string representation of this event.
    */
   public String toString()
   {
      if (year != -1)
      {
         return "Event (" + name + ") at " + new Date(eventTime);
      }
      StringBuffer sb = new StringBuffer("Event (" + name + ") params");
      sb.append(" minute=");
      sb.append(arrToString(minutes));
      sb.append(" hour=");
      sb.append(arrToString(hours));
      sb.append(" dayOfMonth=");
      sb.append(arrToString(daysOfMonth));
      sb.append(" month=");
      sb.append(arrToString(months));
      sb.append(" dayOfWeek=");
      sb.append(arrToString(daysOfWeek));
      sb.append(" (next event date=" + new Date(eventTime) + ")");
      return sb.toString();
   }
}
