package biz.capitalware.mqfm.schedule;

import biz.capitalware.mqfm.MQFM;

/**
 * This class manages the thread which sleeps until the next event. Methods are
 * synchronized to prevent interference from the EventWaiter thread and external
 * threads.
 * 
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ScheduleWaiter implements Runnable
{
   protected ScheduleManager  mgr;
   protected Thread           thread;
   private long               sleepUntil = -1;
   private boolean            shutdown = false;

   /**
    * Creates a new EventWaiter.
    * 
    * @param isDaemon
    *           true if the waiter thread should run as a daemon.
    * @param threadName
    *           the name of the waiter thread
    */
   public ScheduleWaiter(ScheduleManager mgr, boolean isDaemon, String waiterName)
   {
      this.mgr = mgr;

      // start the thread
      thread = new Thread(this, waiterName);
      thread.setPriority(1);
      thread.setDaemon(isDaemon);
      thread.start();
   }

   /**
    * Updates the time to sleep.
    * 
    * @param _sleep_until
    *           the new time to sleep until.
    */
   public synchronized void update(long _sleep_until)
   {
      this.sleepUntil = _sleep_until;
      MQFM.logger.debug("Update for " + _sleep_until); // timeToSleep);
      MQFM.logger.debug("calling notify() to update thread wait timeout");
      notify();
   }

   /**
    * Restarts the thread for a new time to sleep until.
    * 
    * @param _sleep_until
    *           the new time to sleep until.
    */
   public synchronized void restart(long _sleep_until)
   {
      this.sleepUntil = _sleep_until;
      notify();
   }

   /**
    * Stops (destroy) the thread.
    */
   public synchronized void stop()
   {
      shutdown = true;
      notify();
   }

   public synchronized void run()
   {
      MQFM.logger.debug("running");
      while (!shutdown)
      {
         try
         {
            // check if there's an event scheduled
            if (sleepUntil <= 0)
            {
               // no event. Wait for a new event to come along.
               wait();
            } // if
            else
            {
               // Found event, set timeout based on event time
               long timeout = sleepUntil - System.currentTimeMillis();
               if (timeout > 0)
               {
                  wait(timeout);
               }
            }

            // now that we've awakened again, check if an event is due (within
            // 1 second or already past)
            if (sleepUntil >= 0
                  && (sleepUntil - System.currentTimeMillis() < 1000))
            {
               // yes, an event is ready (or already past). Notify the manager
               // to ring it.
               sleepUntil = -1;
               MQFM.logger.debug("notifying manager to ring next event");
               mgr.notifyNextSchedule();
            }
            else
            {
               MQFM.logger.debug("awakened but nothing to do");
            }

         }
         catch (InterruptedException e)
         {
            MQFM.logger.debug("interrupted");
         }
      }
      MQFM.logger.debug("stopping");
   }
}
