package biz.capitalware.mqfm.watch;

import java.util.*;
import java.io.*;

import biz.capitalware.mqfm.handlers.DirectoryWatchFilter;
import biz.capitalware.mqfm.MQFM;

/**
 * Watch Class that watches a directory and all files that follow the filter
 * (extension) criterion
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class DirectoryWatch extends Watch
{
   private String                mainFilename;
   private HashMap<String,Long>  fileList;
   private String[]              filter;
   private FileFilter            dirWatchFilter;
   private File                  file;

   /**
    * @param name
    *           Name of that watch (informal, description)
    * @param directory
    *           directory name where the files to watch are located
    * @param filter
    *           String array containing the extensions of filetypes to be
    *           watched if * is among them, all files are watched
    * @param sendOnStartup
    *           Send existing matching files on startup
    */
   public DirectoryWatch(String directory, String[] filter, boolean sendOnStartup)
   {
      super(directory);
      this.filter = filter;
      mainFilename = directory;
      // instance of FileFilter that accepts only files with extension in filter[]
      dirWatchFilter = new DirectoryWatchFilter(filter);
      // main file object
      file = new File(directory);
      fileList = new HashMap<String,Long>(100);

      if (!sendOnStartup)
      {
         File[] files = file.listFiles(dirWatchFilter);
         // build initial filelist hashmap out of files in the filename directory
         // that fit to the filter criterion
         for (int i = 0; i < files.length; i++)
         {
            String s = files[i].getName();
            fileList.put(s, new Long(files[i].lastModified()));
         }
      }
   }

   /**
    * Checks the current directory if modifications to files occurred.
    *
    * @return null if no changes are detected; Modification String if changes
    *         are detected
    */
   public ArrayList<String> check()
   {
      ArrayList<String> al = new ArrayList<String>();
      // list of current files in the watch directory
      // check if files have been modified or added
      File[] files = file.listFiles(dirWatchFilter);
      for (int i = 0; i < files.length; i++)
      {
         String currentName = files[i].getName();

         // get Date of "old" file
         Long oldFileLong = (Long) fileList.get(currentName);
         long oldFileLastModified = 0;

         if (oldFileLong != null)
            oldFileLastModified = oldFileLong.longValue();

         // file not in list? new created?
         if (oldFileLong == null)
         {
            fileList.put(currentName, new Long(files[i].lastModified()));
            al.add(files[i].getAbsolutePath());
         }
         else
         // file exists in list, but has been modified?
         if (oldFileLastModified != files[i].lastModified())
         {
            fileList.put(currentName, new Long(files[i].lastModified()));
            al.add(files[i].getAbsolutePath());
         }
      }

      // check if files have been deleted - remove it from the list.
      Iterator<String> it = fileList.keySet().iterator();
      while (it.hasNext())
      {
         String fn = mainFilename + File.separator + it.next();
         File f = new File(fn);
         // file deleted ?!
         if (!f.exists())
         {
            it.remove();
         }
      }

      return al;
   }

   /**
    * Output the values used by this class.
    */
   public void printValues()
   {
      MQFM.logger.info("     Directory to be monitored = " + mainFilename);
      for (int i=0; i < filter.length; i++)
         MQFM.logger.info("     Extension to be monitored = " + filter[i]);
   }

   /**
    * Reset the last modified time stamp
    */
   public void resetLastModified(String filename)
   {
      File f = new File(filename);
      String currentName = f.getName();
 
      fileList.put(currentName, new Long(0));
   }
}