package biz.capitalware.mqfm.watch;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.IOException;
import java.util.ArrayList;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.action.objects.SendObject;
import biz.capitalware.mqfm.handlers.IO;
import biz.capitalware.mqfm.handlers.Tokens;
import biz.capitalware.mqfm.mq.MQPutFile;
import biz.capitalware.mqfm.objects.MQFMException;
import biz.capitalware.mqfm.other.GeneralInfo;

import com.ibm.mq.MQException;

/**
 * TimerListener implements an ActionListener interface and listens to the
 * main timer. Each time the timer fires an action Event, all defined watches
 * are checked and the desired file is put to the queue.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class TimerListener implements ActionListener
{
   private Watch[]      watch;
   private boolean      working = false;

   /**
    * @param watch
    *           Array of watches containing the watch information as Watch
    *           objects
    * @param mailinfo
    *           mail related information
    */
   public TimerListener(Watch[] watch)
   {
      super();
      this.watch = watch;
   }

   /**
    * <P>
    * This method is called from the Timer instance (MQFM) every time
    * the polling interval has been finished.
    * </P>
    *
    * <P>
    * Then every time the actionEvent is fired, all defined watches are checked
    * and if a match is found, push the file.
    * </P>
    */
   public void actionPerformed(ActionEvent ae)
   {
      String watchFile = null;
      MQFM.logger.debug("TimerListener: actionPerformed");
      
      working = false;
      for (int i = 0; i < watch.length; i++)
      {
         ArrayList<String> al = watch[i].check();
         if (al.size() > 0)
         {
            MQFM.logger.info("There is/are "+al.size()+" item(s) to send.");
            working = true;
            MQPutFile mqh = null;

            try
            {
               mqh = new MQPutFile(watch[i].getMQObject());

               mqh.connect();

               for (int j=0; j < al.size(); j++)
               {
                  watchFile = al.get(j);
                  MQFM.logger.info("Watched file found: " + watchFile);
                  SendObject so = createSendObject(watchFile, watch[i].getSendObject());
                  mqh.simplePut1(so);
                  if ( (so.archiveFile) || (so.deleteFile) )
                  {
                     resetLastModified(watchFile, i, al);                     
                  }
               }
            }
            catch (IOException e)
            {
               GeneralInfo.DumpHeader();
               MQFM.logger.fatal("IOException: " + e.getLocalizedMessage());
               if (e.getCause() != null)
                  MQFM.logger.fatal(e.getCause().toString());
               resetLastModified(watchFile, i, al);
            }
            catch (MQException e)
            {
               MQFM.logger.fatal("MQException: " + e.getLocalizedMessage());
               if (e.getCause() != null)
                  MQFM.logger.fatal("MQException: " + e.getCause());
               MQFM.logger.fatal("MQCC=" + e.completionCode + " : MQRC="+e.reasonCode);
               resetLastModified(watchFile, i, al);
            }
            catch (MQFMException e)
            {
               MQFM.logger.fatal("MQFMException: " + e.getLocalizedMessage());
               resetLastModified(watchFile, i, al);
            }
            catch (Exception e)
            {
               MQFM.logger.fatal(e.getLocalizedMessage());
               if (e.getCause() != null)
                  MQFM.logger.fatal(e.getCause().toString());
               watch[i].resetLastModified(watchFile);
               resetLastModified(watchFile, i, al);
            }
            finally
            {
               if ( mqh != null)
                  mqh.disconnect();
               MQFM.logger.debug("TimerListener: actionPerformed: mqh.disconnect()");
            }
            
            MQFM.logger.debug("TimerListener: actionPerformed: after try/catch");
         }
      }
      working = false;
      MQFM.logger.debug("TimerListener: actionPerformed: end");
   }

   /**
    * Error condition - handle resetting of last modified time stamp
    * @param watchFile
    * @param i
    * @param al
    */
   public void resetLastModified(String watchFile, int i, ArrayList<String> al)
   {
      MQFM.logger.debug("TimerListener: resetLastModified");
      
      if (watchFile == null)
      {
         for (int j=0; j < al.size(); j++)
            watch[i].resetLastModified(al.get(j));
      }
      else
         watch[i].resetLastModified(watchFile);
   }

   /**
    * Are in the process of sending files?
    * @return true / false
    */
   public boolean isWorking()
   {
      return working;
   }

   /**
    * Create the SendObject from the values in the WatchObject
    * @param pathAndFilename
    * @return
    */
   private SendObject createSendObject(String pathAndFilename, SendObject origSO)
   {
      SendObject so = new SendObject();

      so.actionName = "Watch";
      so.srcFilename = pathAndFilename;

      so.mqXMLFile = origSO.mqXMLFile;
      so.CCDTFile = origSO.CCDTFile;
      so.qMgrName = origSO.qMgrName;
      so.qObject = origSO.qObject;
      so.format = origSO.format;
      so.jobXmlFile = origSO.jobXmlFile;
      so.remoteJobXmlFile = origSO.remoteJobXmlFile;
      so.remoteDirectory = origSO.remoteDirectory;
      so.remoteFilename = origSO.remoteFilename;
      so.archiveDirectory = origSO.archiveDirectory;
      so.archiveFileName = origSO.archiveFileName;
      so.priority = origSO.priority;
      so.persistence = origSO.persistence;
      so.deleteFile = origSO.deleteFile;
      so.archiveFile = origSO.archiveFile;
      so.archiveCreateDir = origSO.archiveCreateDir;
      so.crypto = origSO.crypto;

      if (so.remoteFilename == null)
         so.remoteFilename = IO.getFileName(pathAndFilename);

      if (so.srcFilename != null)
         so.srcFilename = Tokens.replace(so.srcFilename, MQFM.userProperties);
      if (so.remoteFilename != null)
         so.remoteFilename = Tokens.replace(so.remoteFilename, MQFM.userProperties);

      if (so.archiveDirectory != null)
         so.archiveDirectory = Tokens.replace(so.archiveDirectory, MQFM.userProperties);
      if (so.archiveFileName != null)
         so.archiveFileName = Tokens.replace(so.archiveFileName, MQFM.userProperties);

      return so;
   }
}
