package biz.capitalware.mqfm.xml;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

import org.apache.log4j.Appender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Logger;

import org.jdom.*;
import org.jdom.input.*;

import com.ibm.mq.MQC;

import biz.capitalware.mqfm.MQFM;
import biz.capitalware.mqfm.action.objects.*;
import biz.capitalware.mqfm.handlers.IO;
import biz.capitalware.mqfm.handlers.StringUtilities;
import biz.capitalware.mqfm.handlers.Tokens;
import biz.capitalware.mqfm.mq.objects.QueueObject;
import biz.capitalware.mqfm.objects.Constants;
import biz.capitalware.mqfm.objects.Cryptography;

/**
 * This class handles the reading and parsing of MQFM XML file and parse parameters
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class MQFMXml extends FMXml
{
   private ArrayList<FMObject>      _actionAL;
   private Hashtable<String,String> _properties =  new Hashtable<String,String>();
   private OnErrorObject            _onError = null;
   //
   private String                   mqXMLFile = null;
   private String                   CCDTFile = null;
   private String                   qMgrName = null;
   private QueueObject              qObject = null;
   //
   private String                   defaultMQXmlFile = null;
   private String                   defaultCCDTFile = null;
   private String                   defaultQMgrName = null;
   private QueueObject              defaultQObject = null;
   private QueueObject              defaultBackOutQObject = null;
   
   private boolean                  defaultOnErrorFail = true;
   
   private Logger                   logger;

   /**
    * The constructor
    *
    * @param logger
    */
   public MQFMXml()
   {
      super();
   }

   /**
    * Set the Log4J logger
    * @param logger
    */
   public void setLogger(Logger logger)
   {
      this.logger = logger;
   }
   
   /**
    * Read xml file and set parameters
    *
    * @param xmlFilename
    *           The name of xml file
    * @throws JDOMException, IOException
    */
   public void readXMLFile(String xmlFilename) throws JDOMException, IOException
   {
      SAXBuilder builder = new SAXBuilder(true);
      Document doc = builder.build(xmlFilename);

      Element rootElement = doc.getRootElement();

      if ( rootElement.getChild(Constants.GLOBAL_ELEMENT) != null)
         retrieveGlobal(rootElement);

      Element actionsElement = rootElement.getChild(Constants.ACTIONS_ELEMENT);
      
      _actionAL = extractActions(actionsElement);

      addServerSettings();
   }
   
   /**
    * 
    *
    * @param xmlFilename
    *           The name of xml file
    * @throws JDOMException, IOException
    */
   public ArrayList<FMObject> extractActions(Element actionsElement) throws JDOMException
   {
      ArrayList<FMObject> al = new ArrayList<FMObject>();

      Iterator it = actionsElement.getChildren().iterator();

      while (it.hasNext())
      {
         Element actionElement = (Element) it.next();
         String actionTagName = actionElement.getName();

         if (Constants.SEND_ELEMENT.equals(actionTagName))
            al.add(retrieveSend(actionElement));
         else if  (Constants.WATCH_ELEMENT.equals(actionTagName))
            al.add(retrieveWatch( actionElement));
         else if  (Constants.RECEIVE_ELEMENT.equals(actionTagName))
            al.add(retrieveReceive(actionElement));
         else if  (Constants.PUTQUIT_ELEMENT.equals(actionTagName))
            al.add(retrievePutQuit(actionElement));
         else if  (Constants.EXECUTE_ELEMENT.equals(actionTagName))
            al.add(retrieveExecute(actionElement));
         else if  (Constants.IF_ELEMENT.equals(actionTagName))
            al.add(retrieveIf(actionElement));
         else if  (Constants.LAUNCH_ELEMENT.equals(actionTagName))
            al.add(retrieveLaunch(actionElement));
         else if  (Constants.SCHEDULE_ELEMENT.equals(actionTagName))
            al.add(retrieveSchedule(actionElement));
         else if  (Constants.SEND_EMAIL_ELEMENT.equals(actionTagName))
            al.add(retrieveSendEmail(actionElement));
         else if  (Constants.SLEEP_ELEMENT.equals(actionTagName))
            al.add(retrieveSleep(actionElement));
         else if  (Constants.APPEND_ELEMENT.equals(actionTagName))
            al.add(retrieveAppend(actionElement));
         else if  (Constants.COPY_ELEMENT.equals(actionTagName))
            al.add(retrieveCopy(actionElement));
         else if  (Constants.DECRYPT_FILE_ELEMENT.equals(actionTagName))
            al.add(retrieveDecryptFile(actionElement));
         else if  (Constants.DELETE_ELEMENT.equals(actionTagName))
            al.add(retrieveDelete(actionElement));
         else if  (Constants.ENCRYPT_FILE_ELEMENT.equals(actionTagName))
            al.add(retrieveEncryptFile(actionElement));
         else if  (Constants.MERGE_ELEMENT.equals(actionTagName))
            al.add(retrieveMerge(actionElement));
         else if  (Constants.MERGE_SORT_ELEMENT.equals(actionTagName))
            al.add(retrieveMergeSort(actionElement));
         else if  (Constants.MOVE_ELEMENT.equals(actionTagName))
            al.add(retrieveMove(actionElement));
         else if  (Constants.RENAME_ELEMENT.equals(actionTagName))
            al.add(retrieveRename(actionElement));
         else if  (Constants.REPLACE_TEXT_ELEMENT.equals(actionTagName))
            al.add(retrieveReplaceText(actionElement));
         else if  (Constants.SORT_ELEMENT.equals(actionTagName))
            al.add(retrieveSort(actionElement));
         else if  (Constants.TAR_ELEMENT.equals(actionTagName))
            al.add(retrieveTar(actionElement));
         else if  (Constants.TOUCH_ELEMENT.equals(actionTagName))
            al.add(retrieveTouch(actionElement));
         else if  (Constants.UNTAR_ELEMENT.equals(actionTagName))
            al.add(retrieveUnTar(actionElement));
         else if  (Constants.UNZIP_ELEMENT.equals(actionTagName))
            al.add(retrieveUnZip(actionElement));
         else if  (Constants.ZIP_ELEMENT.equals(actionTagName))
            al.add(retrieveZip(actionElement));
         else
            throw new JDOMException("Unknown Action tag '"+actionTagName+"'");
      }
      
      return al;
   }

   /**
    * Retrieve all XML values for Global tag.
    * @param rootElement
    * @throws JDOMException
    */
   private void retrieveGlobal(Element rootElement) throws JDOMException
   {
      Element gblElement = rootElement.getChild(Constants.GLOBAL_ELEMENT);
      Element mqElement = null;
      Element ofElement = null;
      Element execElement = null;
      Element queueElement = null;
      char o = ' ';

      /**
       * Get attributes first.
       */
      defaultOnErrorFail = getOnErrorFailAttribute(gblElement, defaultOnErrorFail);

      if (gblElement.getAttribute(Constants.USE_FILE_LOCKING) != null)
         o = gblElement.getAttribute(Constants.USE_FILE_LOCKING).getValue().toUpperCase().charAt(0);

      if (o == 'Y')
         MQFM.useFileLocking = true;
      else if (o == 'N')
         MQFM.useFileLocking = false;

      /**
       * now get elements
       */
      
      /* Do we have an LogFile element? */
      if ( gblElement.getChild(Constants.LOGFILE_ELEMENT) != null)
      {
         String logfile = Tokens.replace(gblElement.getChild(Constants.LOGFILE_ELEMENT).getTextTrim(), _properties);
         
         Appender appender = logger.getAppender(Constants.LOG4J_APPENDER_MQFM);
         if ( logfile.indexOf(File.separator) == -1 )
            ((FileAppender) appender).setFile("log"+File.separator+logfile);
         else
            ((FileAppender) appender).setFile(logfile);

         ((FileAppender) appender).activateOptions();
      }
      
      /* Do we have a Property elements? */
      List propertyList = gblElement.getChildren(Constants.PROPERTY_ELEMENT);
      Iterator it = propertyList.iterator();
      while (it.hasNext())
      {
         String name = null;
         String value = null;
         Element propElement = (Element) it.next();

         if (propElement.getAttribute(Constants.NAME_ATTRIBUTE) != null)
            name = propElement.getAttribute(Constants.NAME_ATTRIBUTE).getValue().trim();

         if (propElement.getAttribute(Constants.VALUE_ATTRIBUTE) != null)
            value = propElement.getAttribute(Constants.VALUE_ATTRIBUTE).getValue().trim();

         if ( (name != null) && (value != null) && (!("".equals(name))) && (!("".equals(value))) )
            _properties.put(("${"+name+"}"), Tokens.replace(value, _properties));
      }

      /* Do we have an OnFailure element? */
      if ((ofElement = gblElement.getChild(Constants.ON_ERROR_ELEMENT)) != null)
      {
         if ((execElement = ofElement.getChild(Constants.EXECUTE_ELEMENT)) != null)
         {
            if (_onError == null)
               _onError = new OnErrorObject();
            
            _onError.eo = new ExecuteObject();

            if (execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
               _onError.eo.xmlFile = Tokens.replace(execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim(), _properties);
         }
         
         if ((execElement = ofElement.getChild(Constants.SEND_EMAIL_ELEMENT)) != null)
         {
            if (_onError == null)
               _onError = new OnErrorObject();
            
            _onError.seo = new SendEmailObject();

            if (execElement.getAttribute(Constants.FILE_ATTRIBUTE) != null)
               _onError.seo.xmlFile = Tokens.replace(execElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim(), _properties);
         }
      }

      /* Do we have an MQ element? */
      if ((mqElement = gblElement.getChild(Constants.MQ_ELEMENT)) != null)
      {
         if ( mqElement.getChild(Constants.MQFILE_ELEMENT) != null)
            defaultMQXmlFile = mqElement.getChild(Constants.MQFILE_ELEMENT).getTextTrim();

         if ( mqElement.getChild(Constants.CCDTFILE_ELEMENT) != null)
            defaultCCDTFile = mqElement.getChild(Constants.CCDTFILE_ELEMENT).getTextTrim();

         if ( (defaultMQXmlFile != null) && (defaultCCDTFile != null) )
            throw new JDOMException("Error: Do not specify both "+Constants.MQFILE_ELEMENT+" and "+Constants.CCDTFILE_ELEMENT+" elements.");

         if ( mqElement.getChild(Constants.QMGRNAME_ELEMENT) != null)
            defaultQMgrName = mqElement.getChild(Constants.QMGRNAME_ELEMENT).getTextTrim();
         
         if ( (queueElement = mqElement.getChild(Constants.QUEUENAME_ELEMENT)) != null)
         {
            defaultQObject = new QueueObject();
            defaultQObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE) != null)
               defaultQObject.qMgrName = queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               defaultQObject.qMgrName = null;
         }
         
         if ( (queueElement = mqElement.getChild(Constants.BACKOUTQNAME_ELEMENT)) != null)
         {
            defaultBackOutQObject = new QueueObject();
            defaultBackOutQObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE) != null)
               defaultBackOutQObject.qMgrName = queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               defaultBackOutQObject.qMgrName = null;
         }
      }
   }

   /**
    * Retrieve all XML values for send action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveSend(Element actionElement) throws JDOMException
   {
      Element mqElement = retrieveMQ(actionElement);
      Element fileElement = null;
      Element execElement = null;
      Element archiveElement = null;
      String  flag = " ";

      SendObject so = new SendObject();
      so.mqXMLFile = mqXMLFile;
      so.CCDTFile = CCDTFile;
      so.qMgrName = qMgrName;
      so.qObject = qObject;
      
      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.PRIORITY_ATTRIBUTE) != null)
      {
         String priority = actionElement.getAttribute(Constants.PRIORITY_ATTRIBUTE).getValue().trim();
         so.priority = StringUtilities.stringToInt(priority, MQC.MQPRI_PRIORITY_AS_Q_DEF);
      }

      so.format = actionElement.getAttribute(Constants.FORMAT_ATTRIBUTE).getValue().toUpperCase().charAt(0);

      if (actionElement.getAttribute(Constants.PERSISTENCE_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         so.persistence = true;
      else
         so.persistence = false;

      if (actionElement.getAttribute(Constants.DELETE_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         so.deleteFile = true;

      if (actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          so.crypto = new Cryptography();
          so.crypto.passPhrase = actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE) != null)
         {
             String keySize = actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE).getValue().trim();
             so.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
         }
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            so.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            so.srcFilename = filename;
      }

      if ((execElement = actionElement.getChild(Constants.EXECUTE_ELEMENT)) != null)
      {
         if (execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
            so.jobXmlFile = execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();
      }

      if ((archiveElement = actionElement.getChild(Constants.ARCHIVE_ELEMENT)) != null)
      {
         if (archiveElement.getAttribute(Constants.TO_DIR_ATTRIBUTE) != null)
            so.archiveDirectory = archiveElement.getAttribute(Constants.TO_DIR_ATTRIBUTE).getValue().trim();

         if (archiveElement.getAttribute(Constants.TO_FILE_ATTRIBUTE) != null)
            so.archiveFileName = archiveElement.getAttribute(Constants.TO_FILE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
         {
            flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
            if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
               so.archiveCreateDir = true;
         }

         so.archiveFile = true;
      }

      if ( actionElement.getChild(Constants.REMOTE_ELEMENT) != null)
      {
         Element remoteElement = actionElement.getChild(Constants.REMOTE_ELEMENT);

         if ((execElement = remoteElement.getChild(Constants.EXECUTE_ELEMENT)) != null)
         {
            if (execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
               so.remoteJobXmlFile = execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();
         }

         if ( remoteElement.getChild(Constants.DIRECTORY_ELEMENT) != null)
            so.remoteDirectory = remoteElement.getChild(Constants.DIRECTORY_ELEMENT).getTextTrim();

         if ( remoteElement.getChild(Constants.FILENAME_ELEMENT) != null)
            so.remoteFilename = remoteElement.getChild(Constants.FILENAME_ELEMENT).getTextTrim();
      }

      return so;
   }

   /**
    * Retrieve all XML values for watch action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveWatch(Element actionElement) throws JDOMException
   {
      Element mqElement = retrieveMQ(actionElement);
      Element execElement = null; 
      Element archiveElement = null;
      Element remoteElement = null;
      String  flag = " ";
         
      WatchObject wo = new WatchObject();
      wo.mqXMLFile = mqXMLFile;
      wo.CCDTFile = CCDTFile;
      wo.qMgrName = qMgrName;
      wo.qObject = qObject;
      
      wo.labelName = getActionLabelAttribute(actionElement);

      wo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.PRIORITY_ATTRIBUTE) != null)
      {
         String priority = actionElement.getAttribute(Constants.PRIORITY_ATTRIBUTE).getValue().trim();
         wo.priority = StringUtilities.stringToInt(priority, MQC.MQPRI_PRIORITY_AS_Q_DEF);
      }

      if (actionElement.getAttribute(Constants.FORMAT_ATTRIBUTE) != null)
         wo.format = actionElement.getAttribute(Constants.FORMAT_ATTRIBUTE).getValue().toUpperCase().charAt(0);

      if (actionElement.getAttribute(Constants.PERSISTENCE_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         wo.persistence = true;
      else
         wo.persistence = false;

      if (actionElement.getAttribute(Constants.DELETE_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         wo.deleteFile = true;

      if (actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          wo.crypto = new Cryptography();
          wo.crypto.passPhrase = actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE) != null)
         {
             String keySize = actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE).getValue().trim();
             wo.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
         }
      }

      if (actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
         wo.xmlFile = actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();

      if ((execElement = actionElement.getChild(Constants.EXECUTE_ELEMENT)) != null)
      {
         if (execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
            wo.jobXmlFile = execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();
      }

      if ((archiveElement = actionElement.getChild(Constants.ARCHIVE_ELEMENT)) != null)
      {
         if (archiveElement.getAttribute(Constants.TO_DIR_ATTRIBUTE) != null)
            wo.archiveDirectory = archiveElement.getAttribute(Constants.TO_DIR_ATTRIBUTE).getValue().trim();

         if (archiveElement.getAttribute(Constants.TO_FILE_ATTRIBUTE) != null)
            wo.archiveFileName = archiveElement.getAttribute(Constants.TO_FILE_ATTRIBUTE).getValue().trim();

         if (archiveElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
         {
            flag = archiveElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
            if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
               wo.archiveCreateDir = true;
         }

         wo.archiveFile = true;
      }

      if ((remoteElement = actionElement.getChild(Constants.REMOTE_ELEMENT)) != null)
      {
         if ((execElement = remoteElement.getChild(Constants.EXECUTE_ELEMENT)) != null)
         {
            if (execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
               wo.jobXmlFile = execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();
         }

         if ( remoteElement.getChild(Constants.DIRECTORY_ELEMENT) != null)
            wo.remoteDirectory = remoteElement.getChild(Constants.DIRECTORY_ELEMENT).getTextTrim();

         if ( remoteElement.getChild(Constants.FILENAME_ELEMENT) != null)
            wo.remoteFilename = remoteElement.getChild(Constants.FILENAME_ELEMENT).getTextTrim();
      }

      return wo;
   }

   /**
    * Retrieve all XML values for receive action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveReceive(Element actionElement) throws JDOMException
   {
      Element mqElement = retrieveMQ(actionElement);
      Element remoteElement = null;
      Element execElement = null;
      Element ddElement = null;
      Element dfnElement = null;
      Element queueElement = null;
      String  run = Constants.ONE_PASS + "";
      String  flag;

      ReceiveObject ro = new ReceiveObject();
      ro.mqXMLFile = mqXMLFile;
      ro.CCDTFile = CCDTFile;
      ro.qMgrName = qMgrName;
      ro.qObject = qObject;
      ro.backOutQName = defaultBackOutQObject;
      
      ro.labelName = getActionLabelAttribute(actionElement);

      ro.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.GETWITHCONVERT_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         ro.getWithConvert = true;
      else
         ro.getWithConvert = false;

      if (actionElement.getAttribute(Constants.RUN_ATTRIBUTE) != null)
         run = actionElement.getAttribute(Constants.RUN_ATTRIBUTE).getValue().toUpperCase();

      if (actionElement.getAttribute(Constants.APPEND_ATTRIBUTE) != null)
      {
         flag = actionElement.getAttribute(Constants.APPEND_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            ro.append = true;
      }

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            ro.createDir = true;
      }

      if (actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          ro.crypto = new Cryptography();
          ro.crypto.passPhrase = actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE) != null)
         {
             String keySize = actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE).getValue().trim();
             ro.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
         }
      }
      
      if ( (mqElement != null) && ((queueElement = mqElement.getChild(Constants.BACKOUTQNAME_ELEMENT)) != null) )
      {
         ro.backOutQName = new QueueObject();         
         ro.backOutQName.qName = queueElement.getTextTrim();

         if (queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE) != null)
            ro.backOutQName.qMgrName = queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE).getValue().trim();
         else
            ro.backOutQName.qMgrName = null;
      }
      
      if ((remoteElement = actionElement.getChild(Constants.DEFAULT_ELEMENT)) != null)
      {
         if ((execElement = remoteElement.getChild(Constants.EXECUTE_ELEMENT)) != null)
         {
            if (execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
               ro.defaultJobXmlFile = execElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();
            
            if (execElement.getAttribute(Constants.OVERRIDE_ATTRIBUTE) != null)
            {
               String djnOverride = execElement.getAttribute(Constants.OVERRIDE_ATTRIBUTE).getValue().trim();
               if ( (djnOverride.charAt(0) == 'Y') || (djnOverride.charAt(0) == 'y') )
                  ro.overrideJobName = true;
            }
         }

         if ((ddElement = remoteElement.getChild(Constants.DIRECTORY_ELEMENT)) != null)
         {
            ro.defaultDirectory = ddElement.getTextTrim();
            
            if (ddElement.getAttribute(Constants.OVERRIDE_ATTRIBUTE) != null)
            {
               String ddOverride = ddElement.getAttribute(Constants.OVERRIDE_ATTRIBUTE).getValue().trim();
               if ( (ddOverride.charAt(0) == 'Y') || (ddOverride.charAt(0) == 'y') )
                  ro.overrideDirectoryName = true;
            }
         }

         if ((dfnElement = remoteElement.getChild(Constants.FILENAME_ELEMENT)) != null)
         {
            ro.defaultFilename = dfnElement.getTextTrim();
            
            if (dfnElement.getAttribute(Constants.OVERRIDE_ATTRIBUTE) != null)
            {
               String dfnOverride = dfnElement.getAttribute(Constants.OVERRIDE_ATTRIBUTE).getValue().trim();
               if ( (dfnOverride.charAt(0) == 'Y') || (dfnOverride.charAt(0) == 'y') )
                  ro.overrideFileName = true;
            }
         }
      }

      if (run.charAt(0) == 'E')
         ro.processType = Constants.RUN_UNTIL_EMPTY;
      else if (run.charAt(0) == 'D')
         ro.processType = Constants.DAEMON;
      else
         ro.processType = Constants.ONE_PASS;

      return ro;
   }

   /**
    * Retrieve all XML values for putquit action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrievePutQuit(Element actionElement) throws JDOMException
   {
      retrieveMQ(actionElement);

      PutQuitObject pqo = new PutQuitObject();
      pqo.mqXMLFile = mqXMLFile;
      pqo.CCDTFile = CCDTFile;
      pqo.qMgrName = qMgrName;
      pqo.qObject = qObject;
      
      pqo.labelName = getActionLabelAttribute(actionElement);

      pqo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      return pqo;
   }

   /**
    * Retrieve all XML values for execute action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveExecute(Element actionElement)
   {
      ExecuteObject eo = new ExecuteObject();

      eo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
         eo.xmlFile = actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();

      return eo;
   }

   /**
    * Retrieve all XML values for if action.
    * @param ifElement
    * @throws JDOMException
    */
   private FMObject retrieveIf(Element ifElement) throws JDOMException
   {
      Element actionsElement;
      Element elseElement;
      IfObject io = new IfObject();
      
      io.onErrorFail = getOnErrorFailAttribute(ifElement, defaultOnErrorFail);

      if (ifElement.getAttribute(Constants.LABEL_NAME_ATTRIBUTE) != null)
         io.labelName = ifElement.getAttribute(Constants.LABEL_NAME_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(Constants.ACTION_NAME_ATTRIBUTE) != null)
         io.ifActionName = ifElement.getAttribute(Constants.ACTION_NAME_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(Constants.FIELD_NAME_ATTRIBUTE) != null)
         io.ifFieldName = ifElement.getAttribute(Constants.FIELD_NAME_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(Constants.OPERATOR_ATTRIBUTE) != null)
         io.operator = ifElement.getAttribute(Constants.OPERATOR_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(Constants.VALUE_ATTRIBUTE) != null)
         io.value = ifElement.getAttribute(Constants.VALUE_ATTRIBUTE).getValue().trim();

      if ( (actionsElement = ifElement.getChild(Constants.ACTIONS_ELEMENT)) != null)
         io.actionAL = extractActions(actionsElement);

      if ( (elseElement = ifElement.getChild(Constants.ELSE_ELEMENT)) != null)
         io.eo = (ElseObject)retrieveElse(elseElement);

      return io;
   }

   /**
    * Retrieve all XML values for else action.
    * @param elseElement
    * @throws JDOMException
    */
   private FMObject retrieveElse(Element elseElement) throws JDOMException
   {
      Element actionsElement;
      ElseObject eo = new ElseObject();

      eo.onErrorFail = getOnErrorFailAttribute(elseElement, defaultOnErrorFail);

      if (elseElement.getAttribute(Constants.EXIT_RC_ATTRIBUTE) != null)
         eo.exitRC = StringUtilities.stringToInt(elseElement.getAttribute(Constants.EXIT_RC_ATTRIBUTE).getValue().trim(), 0);

      if ( (actionsElement = elseElement.getChild(Constants.ACTIONS_ELEMENT)) != null)
         eo.actionAL = extractActions(actionsElement);

      return eo;
   }

   /**
    * Retrieve all XML values for Launch action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveLaunch(Element actionElement)
   {
      LaunchObject so = new LaunchObject();

      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
         so.xmlFile = actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();

      return so;
   }

   /**
    * Retrieve all XML values for Schedule action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveSchedule(Element actionElement)
   {
      ScheduleObject so = new ScheduleObject();

      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
         so.xmlFile = actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();

      return so;
   }

   /**
    * Retrieve all XML values for SendEmail action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveSendEmail(Element actionElement)
   {
      SendEmailObject seo = new SendEmailObject();
      
      seo.labelName = getActionLabelAttribute(actionElement);

      seo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE) != null)
         seo.xmlFile = actionElement.getAttribute(Constants.XML_FILE_ATTRIBUTE).getValue().trim();

      return seo;
   }

   /**
    * Retrieve all XML values for sleep action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveSleep(Element actionElement)
   {
      SleepObject so = new SleepObject();
      
      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.HOURS_ATTRIBUTE) != null)
         so.hours = StringUtilities.stringToInt(actionElement.getAttribute(Constants.HOURS_ATTRIBUTE).getValue().trim(), 0);

      if (actionElement.getAttribute(Constants.MINUTES_ATTRIBUTE) != null)
         so.minutes = StringUtilities.stringToInt(actionElement.getAttribute(Constants.MINUTES_ATTRIBUTE).getValue().trim(), 0);

      if (actionElement.getAttribute(Constants.SECONDS_ATTRIBUTE) != null)
         so.seconds = StringUtilities.stringToInt(actionElement.getAttribute(Constants.SECONDS_ATTRIBUTE).getValue().trim(), 0);

      if (actionElement.getAttribute(Constants.MILLISECONDS_ATTRIBUTE) != null)
         so.milliseconds = StringUtilities.stringToInt(actionElement.getAttribute(Constants.MILLISECONDS_ATTRIBUTE).getValue().trim(), 0);

      return so;
   }

   /**
    * Retrieve all XML values for Append action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveAppend(Element actionElement)
   {
      Element fileElement = null;
      
      AppendObject ao = new AppendObject();
      
      ao.labelName = getActionLabelAttribute(actionElement);

      ao.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            ao.createDir = true;
      }

      ao.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            ao.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            ao.srcFilename = filename;
      }

      return ao;
   }

   /**
    * Retrieve all XML values for copy action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveCopy(Element actionElement)
   {
      Element fileElement = null;
      
      CopyObject co = new CopyObject();
      
      co.labelName = getActionLabelAttribute(actionElement);

      co.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      co.targetDir = actionElement.getAttribute(Constants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.TO_FILE_ATTRIBUTE) != null)
         co.toFilename = actionElement.getAttribute(Constants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            co.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            co.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            co.srcFilename = filename;
      }

      return co;
   }

   /**
    * Retrieve all XML values for DecryptFile action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveDecryptFile(Element actionElement)
   {
      Element fileElement = null;

      DecryptFileObject dfo = new DecryptFileObject();
      
      dfo.labelName = getActionLabelAttribute(actionElement);

      dfo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      dfo.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          dfo.crypto.passPhrase = actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE).getValue().trim();
      }

      if (actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE) != null)
      {
          String keySize = actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE).getValue().trim();
          dfo.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
      }

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            dfo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            dfo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            dfo.srcFilename = filename;
      }

      return dfo;
   }

   /**
    * Retrieve all XML values for delete action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveDelete(Element actionElement)
   {
      DeleteObject dlo = new DeleteObject();
      
      dlo.labelName = getActionLabelAttribute(actionElement);

      dlo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      dlo.filenames = getXMLFileTags(actionElement);
      
      return dlo;
   }

   /**
    * Retrieve all XML values for EncryptFile action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveEncryptFile(Element actionElement)
   {
      Element fileElement = null;

      EncryptFileObject efo = new EncryptFileObject();
      
      efo.labelName = getActionLabelAttribute(actionElement);

      efo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      efo.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          efo.crypto.passPhrase = actionElement.getAttribute(Constants.PASS_PHRASE_ATTRIBUTE).getValue().trim();
      }

      if (actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE) != null)
      {
          String keySize = actionElement.getAttribute(Constants.KEY_SIZE_ATTRIBUTE).getValue().trim();
          efo.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
      }

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            efo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            efo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            efo.srcFilename = filename;
      }

      return efo;
   }

   /**
    * Retrieve all XML values for Merge action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveMerge(Element actionElement)
   {
      MergeObject mo = new MergeObject();
      
      mo.labelName = getActionLabelAttribute(actionElement);

      mo.srcFilenames = getXMLFileTags(actionElement);

      mo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      mo.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            mo.createDir = true;
      }
      
      return mo;
   }

   /**
    * Retrieve all XML values for MergeSort action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveMergeSort(Element actionElement)
   {
      MergeSortObject mso = new MergeSortObject();
      
      mso.labelName = getActionLabelAttribute(actionElement);

      mso.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.COLUMN_ATTRIBUTE) != null)
      {
          String column = actionElement.getAttribute(Constants.COLUMN_ATTRIBUTE).getValue().trim();
          mso.column = StringUtilities.stringToInt(column, 1);
      }
      
      if (actionElement.getAttribute(Constants.ORDER_ATTRIBUTE) != null)
          mso.order = actionElement.getAttribute(Constants.ORDER_ATTRIBUTE).getValue().trim().charAt(0);

      mso.srcFilenames = getXMLFileTags(actionElement);

      mso.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            mso.createDir = true;
      }

      return mso;
   }

   /**
    * Retrieve all XML values for move action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveMove(Element actionElement)
   {
      Element fileElement = null;
      
      MoveObject mo = new MoveObject();
      
      mo.labelName = getActionLabelAttribute(actionElement);

      mo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if ((fileElement =  actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            mo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            mo.srcFilename = filename;
      }

      mo.targetDir = actionElement.getAttribute(Constants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.TO_FILE_ATTRIBUTE) != null)
         mo.toFilename = actionElement.getAttribute(Constants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            mo.createDir = true;
      }

      return mo;
   }

   /**
    * Retrieve all XML values for rename action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveRename(Element actionElement)
   {
      Element fileElement = null;
      
      RenameObject ro = new RenameObject();
      
      ro.labelName = getActionLabelAttribute(actionElement);

      ro.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      ro.toFilename = actionElement.getAttribute(Constants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
         {
            ro.srcFilename   = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
            ro.toFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), ro.toFilename);
         }
         else
            ro.srcFilename = filename;
      }

      return ro;
   }

   /**
    * Retrieve all XML values for ReplaceText action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveReplaceText(Element actionElement)
   {
      Element fileElement = null;
      Element findElement = null;
      Element replaceElement = null;

      ReplaceTextObject rto = new ReplaceTextObject();
      
      rto.labelName = getActionLabelAttribute(actionElement);

      rto.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      rto.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.ROWS_ATTRIBUTE) != null)
      {
         String rows = actionElement.getAttribute(Constants.ROWS_ATTRIBUTE).getValue().trim();
         rto.convertRTValue(rows, rto.row);
      }

      if (actionElement.getAttribute(Constants.COLUMNS_ATTRIBUTE) != null)
      {
          String columns = actionElement.getAttribute(Constants.COLUMNS_ATTRIBUTE).getValue().trim();
          rto.convertRTValue(columns, rto.column);
      }

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            rto.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            rto.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            rto.srcFilename = filename;
      }
      
      if ((findElement = actionElement.getChild(Constants.FIND_ELEMENT)) != null)
         rto.findText = findElement.getText();
      
      if ((replaceElement = actionElement.getChild(Constants.REPLACE_ELEMENT)) != null)
         rto.replaceText = replaceElement.getText();

      return rto;
   }

   /**
    * Retrieve all XML values for Sort action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveSort(Element actionElement)
   {
      Element fileElement = null;
      
      SortObject so = new SortObject();
      
      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      so.toFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.COLUMN_ATTRIBUTE) != null)
      {
          String column = actionElement.getAttribute(Constants.COLUMN_ATTRIBUTE).getValue().trim();
          so.column = StringUtilities.stringToInt(column, 1);
      }
      
      if (actionElement.getAttribute(Constants.ORDER_ATTRIBUTE) != null)
          so.order = actionElement.getAttribute(Constants.ORDER_ATTRIBUTE).getValue().trim().charAt(0);

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            so.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            so.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            so.srcFilename = filename;
      }

      return so;
   }

   /**
    * Retrieve all XML values for Tar action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveTar(Element actionElement)
   {
      Element fileElement = null;
      
      TarObject to = new TarObject();
      
      to.labelName = getActionLabelAttribute(actionElement);

      to.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      to.tarFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            to.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            to.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            to.srcFilename = filename;
      }

      return to;
   }

   /**
    * Retrieve all XML values for touch action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveTouch(Element actionElement)
   {
      TouchObject to = new TouchObject();
      
      to.labelName = getActionLabelAttribute(actionElement);

      to.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            to.createDir = true;
      }

      to.filenames = getXMLFileTags(actionElement);

      return to;
   }

   /**
    * Retrieve all XML values for UnTar action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveUnTar(Element actionElement)
   {
      Element fileElement = null;
      
      UnTarObject uto = new UnTarObject();
      
      uto.labelName = getActionLabelAttribute(actionElement);

      uto.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      uto.targetDir = actionElement.getAttribute(Constants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            uto.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String tarFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            uto.tarFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), tarFilename);
         else
            uto.tarFilename = tarFilename;
      }

      return uto;
   }

   /**
    * Retrieve all XML values for UnZip action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveUnZip(Element actionElement)
   {
      Element fileElement = null;

      UnZipObject uzo = new UnZipObject();
      
      uzo.labelName = getActionLabelAttribute(actionElement);

      uzo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      uzo.targetDir = actionElement.getAttribute(Constants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            uzo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String zipFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            uzo.zipFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), zipFilename);
         else
            uzo.zipFilename = zipFilename;
      }

      return uzo;
   }

   /**
    * Retrieve all XML values for Zip action.
    * @param actionElement
    * @throws JDOMException
    */
   private FMObject retrieveZip(Element actionElement)
   {
      Element fileElement = null;

      ZipObject zo = new ZipObject();
      
      zo.labelName = getActionLabelAttribute(actionElement);

      zo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      zo.zipFilename = actionElement.getAttribute(Constants.FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(Constants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if ( (flag.charAt(0) == 'Y') || (flag.charAt(0) == 'y') )
            zo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(Constants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            zo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            zo.srcFilename = filename;
      }

      return zo;
   }

   /**
    * Retrieve all XML values for MQ tag.
    * @param actionElement
    * @throws JDOMException
    */
   private Element retrieveMQ(Element actionElement) throws JDOMException
   {
      Element mqElement = null;
      Element queueElement = null;
      
      mqXMLFile= defaultMQXmlFile;
      CCDTFile = defaultCCDTFile;
      qMgrName = defaultQMgrName;
      qObject  = defaultQObject;

      if ((mqElement = actionElement.getChild(Constants.MQ_ELEMENT)) != null)
      {
         if ( mqElement.getChild(Constants.MQFILE_ELEMENT) != null)
         {
            mqXMLFile = mqElement.getChild(Constants.MQFILE_ELEMENT).getTextTrim();

            boolean exists = (new File(Constants.MQ_PATH + mqXMLFile)).exists();
            if ( !(exists) )
               throw new JDOMException("Error: The MQ XML file does not exists : " + Constants.MQ_PATH + mqXMLFile);
         }

         if ( mqElement.getChild(Constants.CCDTFILE_ELEMENT) != null)
            CCDTFile = mqElement.getChild(Constants.CCDTFILE_ELEMENT).getTextTrim();

         if ( (mqXMLFile != null) && (CCDTFile != null) )
            throw new JDOMException("Error: Do not specify both "+Constants.MQFILE_ELEMENT+" and "+Constants.CCDTFILE_ELEMENT+" elements.");

         if ( mqElement.getChild(Constants.QMGRNAME_ELEMENT) != null)
            qMgrName = mqElement.getChild(Constants.QMGRNAME_ELEMENT).getTextTrim();
         
         if ( (queueElement = mqElement.getChild(Constants.QUEUENAME_ELEMENT)) != null)
         {
            qObject = new QueueObject();

            qObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE) != null)
               qObject.qMgrName = queueElement.getAttribute(Constants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               qObject.qMgrName = null;
         }
      }

      if (mqXMLFile != null)
         mqXMLFile  = Tokens.replace(mqXMLFile, _properties);
      if (CCDTFile != null)
         CCDTFile   = Tokens.replace(CCDTFile, _properties);
      if (qMgrName != null)
         qMgrName   = Tokens.replace(qMgrName, _properties);
      if (qObject != null)
      {
         if (qObject.qName != null)
            qObject.qName      = Tokens.replace(qObject.qName, _properties);
         if (qObject.qMgrName != null)
            qObject.qMgrName      = Tokens.replace(qObject.qMgrName, _properties);
      }

      return mqElement;
   }

   /**
    * Retrieve 1 or more XML File elements from the section.
    * @param actionElement
    * @return newFilenames
    */
   private String[] getXMLFileTags(Element actionElement)
   {
      Element fileElement = null;
      ArrayList<String> al = new ArrayList<String>();
      String srcFilename = null;

      /* We should have at least 2 File elements */
      List propertyList = actionElement.getChildren(Constants.FILE_ELEMENT);
      Iterator it = propertyList.iterator();
      while (it.hasNext())
      {
         fileElement = (Element) it.next();

         srcFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(Constants.DIR_ATTRIBUTE) != null)
            al.add(IO.apendFileToPath(fileElement.getAttribute(Constants.DIR_ATTRIBUTE).getValue().trim(), srcFilename));
         else
            al.add(srcFilename);
      }
      
      String[] newFilenames = new String[al.size()];
      for (int i=0; i < al.size(); i++)
         newFilenames[i] = al.get(i);
      
      return newFilenames;
   }

   /**
    * Retrieve
    * @param actionElement
    * @return newFilenames
    */
   private String getActionLabelAttribute(Element actionElement)
   {
      String name = "default";

      if (actionElement.getAttribute(Constants.LABEL_ATTRIBUTE) != null)
         name = actionElement.getAttribute(Constants.LABEL_ATTRIBUTE).getValue().trim();

      return name;
   }
   
   /**
    * Add server related tokens and their values to the hashtable.
    */
   private void addServerSettings()
   {
      try
      {
         String value = InetAddress.getLocalHost().getHostName();
         _properties.put(Constants.HOSTNAME_TOKEN, value);
      }
      catch (UnknownHostException e)
      {}

      try
      {
         String value = InetAddress.getLocalHost().getHostAddress();
         _properties.put(Constants.IPADDRESS_TOKEN, value);
      }
      catch (UnknownHostException e)
      {}
   }

   /**
    * Retrieve an ArrayList of all of the actions to be performed.
    * @return actionAL
    */
   public ArrayList<FMObject> getActions()
   {
      return _actionAL;
   }

   /**
    * Retrieve an OnErrorObject.
    * @return onError
    */
   public OnErrorObject getOnError()
   {
      return _onError;
   }

   /**
    * Retrieve user properties.
    * @return properties
    */
   public Hashtable<String,String>  getProperties()
   {
      return _properties;
   }
}
